"""Model training using the Masterful platform. """
import dataclasses
import tensorflow as tf
from typing import Dict, Optional, Sequence, Tuple


@dataclasses.dataclass
class TrainingReport:
  """Structure which holds the results of a training run.

  Args:
    validation_results: The final results from evaluating the model
      on the validation set.
    gpu_info: A list of GpuInfo objects, with relevant gpu usage information.
    history: The full training history report, containing the results at the
      end of each epoch for key metrics.
    model: A reference to the trained model. This will be different
      than the model passed in for training if `model_ensemble` is
      greater than 1.
  """
  validation_results: Dict[str, float] = None
  gpu_info: Sequence["masterful.utils.gpu.GpuInfo"] = None
  history: tf.keras.callbacks.History = None
  model: Optional[tf.keras.Model] = None


def train(
    model: tf.keras.Model,
    model_params: "masterful.architecture.ArchitectureParams",
    optimization_params: "masterful.optimization.OptimizationParams",
    regularization_params: "masterful.regularization.RegularizationParams",
    ssl_params: "masterful.ssl.SemiSupervisedParams",
    training_dataset: tf.data.Dataset,
    training_dataset_params: "masterful.data.DataParams",
    validation_dataset: Optional[tf.data.Dataset] = None,
    validation_dataset_params: Optional["masterful.data.DataParams"] = None,
    unlabeled_datasets: Optional[Sequence[Tuple[
        tf.data.Dataset, "masterful.data.DataParams"]]] = None,
    synthetic_datasets: Optional[Sequence[Tuple[
        tf.data.Dataset, "masterful.data.DataParams"]]] = None,
    **kwargs) -> TrainingReport:
  """Trains a model using the Masterful platform.

  The model passed into this function will be trained against
  the passed in datasets using the given parameters for
  regularization, optimization, and semi-supervised learning.

  Example:

  .. code-block:: python

    model: tf.keras.Model = ...
    model_params: masterful.architecture.params.ArchitectureParams = ...
    optimization_params: masterful.optimization.params.OptimizationParams = ...
    regularization_params: masterful.regularization.RegularizationParams = ...
    ssl_params: masterful.ssl.params.SemiSupervisedParams = ...
    training_dataset: tf.data.Dataset = ...
    training_dataset_params: masterful.data.params.DataParams = ...
    training_report = masterful.training.train(
        model=model,
        model_params=model_params,
        optimization_params=optimization_params,
        regularization_params=regularization_params,
        ssl_params=ssl_params,
        training_dataset=training_dataset,
        training_dataset_params=training_dataset_params)

  Args:
    model: The model to train.
    model_params: Parameters of the model to train.
    optimizer_params: Parameters to use for optimization. These
      can be created directly, or found automatically using
      :func:`masterful.optimization.learn_optimization_params`.
    regularization_params: Parameters to use for regularization.
      These can be created directly, or found automatically using
      :func:`masterful.regularization.learn_regularization_params`.
    ssl_params: Parameters to use for semi-supervised training.
      These can be created directly, or learned automatically
      using :func:`masterful.ssl.learn_ssl_params`.
    training_dataset: The labeled dataset to use during training.
    training_dataset_params: The parameters of the labeled dataset.
    validation_dataset: An optional validation dataset to use during
      training. If no validation set is specified, Masterful will
      autmoatically create one from the labeled dataset.
    validation_dataset_params: Optional parameters of the validation dataset.
    unlabeled_datasets: Optional sequence of unlabled datasets and their
      parameters, to use during training. If an unlabeled dataset is
      specified, then a set of algorithms must be specified in `ssl_params`
      otherwise this will have no effect.
    synthetic_datasets: Optional sequence of synthetic data and parameters
      to use during training. The amount of synthetic data used during
      training is controlled by
      :class:`masterful.regularization.RegularizationParams.synthetic_proportion`.

  Returns:
    An instance of :class:`TrainingReport` with the full results
    of training the model with the given parameters.
  """
  raise RuntimeError(
      "Please call masterful.register() with your account ID and authorization key before using the API."
  )


def _ensemble(
    model: tf.keras.Model,
    model_params: "masterful.architecture.ArchitectureParams",
    optimization_params: "masterful.optimization.OptimizationParams",
    regularization_params: "masterful.regularization.RegularizationParams",
    ssl_params: "masterful.ssl.SemiSupervisedParams",
    training_dataset: tf.data.Dataset,
    training_dataset_params: "masterful.data.DataParams",
    validation_dataset: Optional[tf.data.Dataset] = None,
    validation_dataset_params: Optional["masterful.data.DataParams"] = None,
    unlabeled_datasets: Optional[Sequence[Tuple[
        tf.data.Dataset, "masterful.data.DataParams"]]] = None,
    synthetic_datasets: Optional[Sequence[Tuple[
        tf.data.Dataset, "masterful.data.DataParams"]]] = None,
    **kwargs,
) -> TrainingReport:
  """
  Similar to :py:func:`masterful.training.train`, trains the given model on the
  provided datasets using the policy provided. However,
  this function trains `model_parms.ensemble_multiplier`
  models in sequence, and returns an ensembled model
  which is the joint prediction from all child models.

  Args:
    model: The model to train.
    model_params: Parameters of the model to train.
    optimizer_params: Parameters to use for optimization. These
      can be created directly, or found automatically using
      :func:`masterful.optimization.learn`.
    regularization_params: Parameters to use for regularization.
      These can be created directly, or found automatically using
      :func:`masterful.regularization.learn`.
    ssl_params: Parameters to use for semi-supervised training.
      These can be created directly, or learned automatically
      using :func:`masterful.ssl.learn`.
    training_dataset: The labeled dataset to use during training.
    training_dataset_params: The parameters of the labeled dataset.
    validation_dataset: An optional validation dataset to use during
      training. If no validation set is specified, Masterful will
      autmoatically create one from the labeled dataset.
    validation_dataset_params: Optional parameters of the validation dataset.
    unlabeled_datasets: Optional sequence of unlabled datasets and their
      parameters, to use during training. If an unlabeled dataset is
      specified, then a set of algorithms must be specified in `ssl_params`
      otherwise this will have no effect.
    synthetic_datasets: Optional sequence of synthetic data and parameters
      to use during training. The amount of synthetic data used during
      training is controlled by
      :class:`masterful.regularization.RegularizationParams.synthetic_proportion`.

  Returns:
    An instance of :class:`TrainingReport` with the full results
    of training the model with the given parameters.
  """
  raise RuntimeError(
      "Please call masterful.register() with your account ID and authorization key before using the API."
  )



"""Parameters and algorithms for semi-supervised learning."""
from typing import Optional, Sequence, Tuple
from dataclasses import dataclass, field

import tensorflow as tf


@dataclass
class SemiSupervisedParams:
  """Parameters which control the semi-supervised learning aspects of Masterful training.

  In this context, semi-supervised learning incorporates self training,
  self-supervised learning, and traditional semi-supervised learning
  (any learning with a combination of labeled and unlabeled data).

  Args:
    algorithms: An optional list of semi-supervised learning algorithms
      to use during training. Can be any combination of
      ["noisy_student", "barlow_twins"]. Defaults to `["noisy_student"]`
  """
  algorithms: Optional[Sequence[str]] = field(
      default_factory=lambda: ["noisy_student"])


def learn_ssl_params(
    training_dataset: tf.data.Dataset,
    training_dataset_params: "masterful.data.DataParams",
    unlabeled_datasets: Optional[Sequence[Tuple[
        tf.data.Dataset, "masterful.data.DataParams"]]] = None,
    synthetic_datasets: Optional[Sequence[Tuple[
        tf.data.Dataset, "masterful.data.DataParams"]]] = None,
) -> SemiSupervisedParams:
  """Learns the optimal set of semi-supervised learning parameters to use during training.

  Example:

  .. code-block:: python

    training_dataset: tf.data.Dataset = ...
    training_dataset_params: masterful.data.params.DataParams = ...
    ssl_params = masterful.ssl.learn_ssl_params(training_dataset,
                                                training_dataset_params)

  Args:
    training_dataset: The labeled dataset to use during training.
    training_dataset_params: The parameters of the labeled dataset.
    unlabeled_datasets: Optional sequence of unlabled datasets and their
      parameters, to use during training. If an unlabeled dataset is
      specified, then a set of algorithms must be specified in `ssl_params`
      otherwise this will have no effect.
    synthetic_datasets: Optional sequence of synthetic data and parameters
      to use during training. The amount of synthetic data used during
      training is controlled by
      :class:`masterful.regularization.RegularizationParams.synthetic_proportion`.
  """
  raise RuntimeError(
      "Please call masterful.register() with your account ID and authorization key before using the API."
  )


def analyze_data_then_save_to(
    model: tf.keras.Model,
    labeled_training_data: tf.data.Dataset,
    unlabeled_training_data: tf.data.Dataset,
    path: str,
) -> None:
  """Analyze labeled and unlabeled data then save intermediate results to disk.

  See the Recipe for Simple SSL guide :doc:/notebooks/guide_simple_ssl
  for more details.

  Example:

  .. code-block:: python

    model: tf.keras.Model = ...
    training_dataset: tf.data.Dataset = ...
    unlabeled_dataset: tf.data.Dataset = ...
    masterful.ssl.analyze_data_then_save_to(
        model, training_dataset, unlabeled_dataset, path='/tmp/ssl')

  Args:
    model: A trained model. The output must be probabilities, in
      other words, your model's final layer should be a softmax
      or sigmoid activation.

      If your model finishes with a `tf.keras.layers.Dense` layer,
      without an activation, then it's said to be 'outputting logits'.
      In that case, typically you'll use a loss function initialized with
      `from_logits=True`. If this describes your model, you can simply
      attach an extra sigmoid or softmax activation to your model and pass
      the new model into this function. You do not need to change your original
      model, loss function, or training loop.
      For example, the model below outputs logits:
      ```python
      m = tf.keras.Sequential([tf.keras.Input((32,32,3)),
                               tf.keras.layers.Dense(10)])
      ```

      To use this model, attach a softmax activation:
      ```
      activated_model = tf.keras.Sequential([m, tf.keras.layers.Softmax()])
      masterful.ssl.save_data(activated_model, ...)
      ```

    architecture_params: Parameters about the model architecture.

    labeled_training_data: Labeled training data as a tf.data.Dataset.
      The data should be batched. Each example should have the
      following structure: `(original_images, original_labels)`.

    labeled_training_data_params: Params that describe the labeled training data.

    unlabeled_training_data: Unlabeled training data as a tf.data.Dataset.
      The data should be batched. Each example should be a tensor of
      images.

    unlabeled_training_data_params: Params that describe the unlabeled training data.

    path: The filepath to save to.

    Raises:
      ValueError:	If the path is empty or malformed.
  """
  raise RuntimeError(
      "Please call masterful.register() with your account ID and authorization key before using the API."
  )


def load_from(path: str,
              unlabeled_weight: Optional[float] = 1.0) -> tf.data.Dataset:
  """Load data from disk into a tf.data.Dataset.

  Please see the Recipe for Simple SSL guide for more details.

  Example:

  .. code-block:: python

    ssl_training_dataset = masterful.ssl.load_from(path='/tmp/ssl')

  Args:
    path: The location on disk to load from.
    unlabeled_weight: A weighting for the unlabeled data.

  Returns:
    A dataset ready to be trained against. The dataset is unbatched. If unlabeled_weight
      is specified, and not set to 1.0, the dataset elements are (image, label, weight),
      where weight is 1.0 for labeled data unlabeled_weight for unlabeled data. Otherwise,
      the dataset elements are (image, label).

  Raises:
    ValueError:	If the path is empty or malformed.
    FileNotFound: If the path does not point to a valid file on disk.
  """
  raise RuntimeError(
      "Please call masterful.register() with your account ID and authorization key before using the API."
  )


def distill(
    source_model: tf.keras.Model,
    source_model_params: "masterful.architecture.ArchitectureParams",
    target_model: tf.keras.Model,
    target_model_params: "masterful.architecture.ArchitectureParams",
    optimization_params: Optional["masterful.optimization.OptimizationParams"],
    training_dataset: Optional[tf.data.Dataset],
    training_dataset_params: Optional["masterful.data.DataParams"],
    unlabeled_datasets: Optional[Sequence[Tuple[
        tf.data.Dataset, "masterful.data.DataParams"]]] = None,
    synthetic_datasets: Optional[Sequence[Tuple[
        tf.data.Dataset, "masterful.data.DataParams"]]] = None,
    **kwargs,
) -> "masterful.training.TrainingReport":
  """Distills the knowledge from `source_model` into `target_model`.

  Example:

  .. code-block:: python

    teacher_model: tf.keras.Model = ...
    teacher_model_params: masterful.architecture.params.ArchitectureParams = ...
    student_model: tf.keras.Model = ...
    student_model_params: masterful.architecture.params.ArchitectureParams = ...
    optimization_params: masterful.optimization.params.OptimizationParams = ...
    training_dataset: tf.data.Dataset = ...
    training_dataset_params: masterful.data.params.DataParams = ...
    training_report = masterful.ssl.distill(
        source_model=teacher_model,
        source_model_params=teacher_model_params,
        target_model=student_model,
        target_model_params=student_model_params,
        optimization_params=optimization_params,
        training_dataset=training_dataset,
        training_dataset_params=training_dataset_params)

  Args:
    source_model: The source model that we are trying to match. This model
      is already trained.
    source_model_params: The parameters of the source model.
    target_model: The target model that we are training.
    target_model_params: The parameters of the target model.
    training_dataset: The labeled data to use for training the model. Labeled
      data must be unbatched, and use the Keras formulation of
      (image, label) for each mini-batch of data.
    training_dataset_params: The parameters of the training dataset.
    unlabeled_datasets: [Optional] A set of unlabeled datasets and their parameters
      which can be used to improve the training of the model through semi-supervised
      and unsupervised techniques.
    synthetic_datasets: [Optional] A set of labeled, synthetic data and their parameters
      that can be used to improve the performance of the model.

  Returns:
    An instance of :class:`masterful.training.TrainingReport`,
    containing the results of distilling `source_model` into `target_model`.
  """
  raise RuntimeError(
      "Please call masterful.register() with your account ID and authorization key before using the API."
  )


def learn_representation(
    model: tf.keras.Model,
    model_params: "masterful.data.DataParams",
    optimization_params: "masterful.optimization.OptimizationParams",
    ssl_params: "masterful.ssl.SemiSupervisedParams",
    training_dataset: Optional[tf.data.Dataset] = None,
    training_dataset_params: Optional["masterful.data.DataParams"] = None,
    validation_dataset: Optional[tf.data.Dataset] = None,
    validation_dataset_params: Optional["masterful.data.DataParams"] = None,
    unlabeled_datasets: Optional[Sequence[Tuple[
        tf.data.Dataset, "masterful.data.DataParams"]]] = None,
    synthetic_datasets: Optional[Sequence[Tuple[
        tf.data.Dataset, "masterful.data.DataParams"]]] = None,
    **kwargs,
) -> "masterful.training.TrainingReport":
  """
  Pretrain the weights of the given model using the provided datasets. The model
  is assumed to be the feature extractor (backbone) of a larger model, so there
  should be no classification heads (softmax output) in the model provided.

  Example:

  .. code-block:: python

    model: tf.keras.Model = ...
    model_params: masterful.architecture.params.ArchitectureParams = ...
    optimization_params: masterful.optimization.params.OptimizationParams = ...
    ssl_params: masterful.ssl.params.SemiSupervisedParams = ...
    training_dataset: tf.data.Dataset = ...
    training_dataset_params: masterful.data.params.DataParams = ...
    validation_dataset: tf.data.Dataset = ...
    validation_dataset_params: masterful.data.params.DataParams = ...
    training_report = masterful.ssl.learn_representation(
        model=model,
        model_params=model_params,
        optimization_params=optimization_params,
        ssl_params=ssl_params,
        training_dataset=training_dataset,
        training_dataset_params=training_dataset_params,
        validation_dataset=validation_dataset,
        validation_dataset_params=validation_dataset_params)

  Args:
    model: The model to pretrain. Models used here should have no classification
      head attached.
    model_params: The parameters of the model to train.
    training_dataset: The labeled data to use for training the model. Labeled
      data must be unbatched, and use the Keras formulation of
      (image, label) for each mini-batch of data.
    training_dataset_params: The parameters of the training dataset.
    validation_dataset: The labeled data to use for validating the model. Labeled
      data must be unbatched, and use the Keras formulation of
      (image, label) for each mini-batch of data.
    validation_dataset_params: The parameters of the validation dataset.
    unlabeled_datasets: [Optional] A set of unlabeled datasets and their parameters
      which can be used to improve the training of the model through semi-supervised
      and unsupervised techniques.
    synthetic_datasets: [Optional] A set of labeled, synthetic data and their parameters
      that can be used to improve the performance of the model.

  Returns:
    An instance of :class:`FitReport`, containing the results of pretraining
    the model. In order to measure the performance of pretraining, a small task
    specific head is temporarily attached and trained at the end to measure the
    performance of the pretraining task.
  """
  raise RuntimeError(
      "Please call masterful.register() with your account ID and authorization key before using the API."
  )



""" Data parameters and algorithms. """
from dataclasses import dataclass
import tensorflow as tf
from typing import Optional, Tuple

import masterful.enums


@dataclass
class DataParams:
  """Parameters describing the datasets used during training.

  These parameters describe both the structure of the dataset
  (image and label shapes for examples) as well as semantic
  structure of the labels (the bounding box format for example,
  or whether or not the labels are sparse or dense).

  Args:
    num_classes: The number of possible classes in the dataset.
    task: The task this dataset will be used for.
    image_shape: The input shape of image data in the dataset,
      in the format (height, width, channels) if `input_channels_last=True`,
      otherwise (channels, height, width) if `input_channels_last=False`.
    image_range: The range of pixels in the input image space that
      of the dataset.
    image_dtype: The image data type in the dataset.
    image_channels_last: The ordering of the dimensions in the inputs.
      `input_channels_last=True` corresponds to inputs with shape
      (height, width, channels) while `input_channels_last=False`
      corresponds to inputs with shape (channels, height, width). Defaults
      to True.
    label_dtype: The data type of the labels.
    label_shape: The shape of the labels.
    label_structure: The tensor format of the label examples.
    label_sparse: True if the labels are in sparse format, False
      for dense (one-hot) labels.
    label_bounding_box_format: The format of bounding boxes in the label,
      if they exist.
  """
  num_classes: int = None
  task: "masterful.enums.Task" = None

  image_shape: Tuple = None
  image_range: "masterful.enums.ImageRange" = None
  image_dtype: tf.dtypes.DType = None
  image_channels_last: bool = True

  label_dtype: type = None
  label_shape: Tuple = None
  label_structure: "masterful.enums.TensorStructure" = None
  label_sparse: bool = None
  label_bounding_box_format: Optional[
      "masterful.enums.BoundingBoxFormat"] = None

  @property
  def image_width(self) -> int:
    """Gets the width of the images in this dataset."""
    if len(self.image_shape) > 2:
      return (self.image_shape[1]
              if self.image_channels_last else self.image_shape[2])
    else:
      return self.image_shape[1]

  @property
  def image_height(self) -> int:
    """Gets the height of the images in this dataset."""
    if len(self.image_shape) > 2:
      return (self.image_shape[0]
              if self.image_channels_last else self.image_shape[1])
    else:
      return self.image_shape[0]

  @property
  def image_channels(self) -> int:
    """Gets the number of channels in the images in this dataset."""
    if len(self.image_shape) > 2:
      return (self.image_shape[2]
              if self.image_channels_last else self.image_shape[0])
    else:
      return 0


def learn_data_params(
    dataset: tf.data.Dataset,
    image_range: "masterful.enums.ImageRange",
    num_classes: int,
    sparse_labels: bool,
    task: "masterful.enums.Task",
    bounding_box_format: "masterful.enums.BoundingBoxFormat" = None,
) -> DataParams:
  """Learns the :class:`DataParams` for the given dataset.

  Most parameters can be introspected from the dataset itself.
  Anything that cannot be introspected is passed into this function
  as an argument, or set on the :class:`DataParams` after creation.

  Example:

  .. code-block:: python

    training_dataset: tf.data.Dataset = ...
    dataset_params = masterful.data.learn_data_params(
        dataset=training_dataset,
        image_range=masterful.enums.ImageRange.ZERO_255,
        num_classes=10,
        sparse_labels=False,
        task=masterful.enums.Task.CLASSIFICATION)

  Args:
    dataset: A `tf.data.Dataset` instance to learn the parameters for.
    image_range: The range of pixels in the input image space that
      of the dataset.
    num_classes: The number of possible classes in the dataset.
    sparse_labels: True if the labels are in sparse format, False
      for dense (one-hot) labels.
    task: The task this dataset will be used for.
    bounding_box_format: The format of bounding boxes in the label,
      if they exist.

  Returns:
    A new instance of DataParams describing the passed in dataset.
  """
  raise RuntimeError(
      "Please call masterful.register() with your account ID and authorization key before using the API."
  )



"""A module responsible for Masterful code deployment."""

import atexit
import json
import os
import time
import sys
import types
import uuid
from datetime import datetime
import re

import memory_tempfile
import requests
from requests import auth

from masterful import constants
from masterful import version


try:
  memory_tempfile = memory_tempfile.MemoryTempfile(fallback=True)
  _DOWNLOAD_FOLDER = memory_tempfile.gettempdir()
except AttributeError:
  import tempfile
  _DOWNLOAD_FOLDER = tempfile.gettempdir()
except RuntimeError:
  import tempfile
  _DOWNLOAD_FOLDER = tempfile.gettempdir()

# Seconds to sleep the register call after 30 days of trial period
_NAGWARE_SLEEP_SECONDS = 30
# Number of trial days where an user can use masterful freely
_MASTERFUL_TRIAL_DAYS = 45
# Number of to filter the nagging message sent to clients
_MASTERFUL_REMINDER_DAYS = 30

_MASTERFUL_TRIAL_ACCOUNT_ID = '631d9db2-39cd-420c-a6aa-4c7c7f81464d'
_MASTERFUL_TRIAL_AUTHORIZATION_KEY = '6cef878d-fc77-4f1f-a2c5-2f3db6b6258a'

_MASTERFUL_REMINDER_DAYS_MESSAGE = (
  'Loaded Masterful version {}. This software is distributed free of\n'
  'charge for personal projects and evaluation purposes.\n'
  'See http://www.masterfulai.com/personal-and-evaluation-agreement for details.\n'
  'Sign up in the next {} days at https://www.masterfulai.com/get-it-now\n'
  'to continue using Masterful.'
)

_MASTERFUL_REMINDER_SIGN_UP_NOW = (
  'Loaded Masterful version {}. This software is distributed free of\n'
  'charge for personal projects and evaluation purposes.\n'
  'See http://www.masterfulai.com/personal-and-evaluation-agreement for details.\n'
  'Sign up now at https://www.masterfulai.com/get-it-now\n'
  'to continue using Masterful.'
)

_MASTERFUL_REMINDER_TRIAL_VERSION_FINISHED = (
  'Unable to load Masterful: sign up required.\n'
  'Sign up now at https://www.masterfulai.com/get-it-now to continue your free\n'
  'usage. Terminating.'
)


def _pretty_error_message(response, account_id) -> str:
  status_code = response.status_code
  account_status = response.json().get('context', {}).get('account_status', '')
  pretty_messages = {
      400:
          f"The Masterful account id '{account_id}' or authorization key is in a bad format. Please register for a new account at https://www.masterfulai.com/try-free.",
      401:
          f"The authorization key for Masterful account id '{account_id}' is incorrect. Please check your key and try again.",
      403:
          f"The Masterful account id '{account_id}' has hit its free usage limits ({account_status}). Please contact your Masterful representative to get a new key.",
      404:
          f"The Masterful account id '{account_id}' does not exist. Please check your key and try again.",
  }
  if status_code in pretty_messages:
    print(ValueError(pretty_messages[status_code]))
  else:
    print(
        f'Failed to register Masterful package. (Internal Server Error: {response.status_code}){os.linesep}'
        f'{json.dumps(response.json(), indent=2)}')



def _print_success_message(version: str) -> None:
  """Prints a user friendly message to confirm the success of register.

  Args:
    version: The package version to include in the message.
  """
  print(
    "MASTERFUL: Your account has been successfully registered."
    f" Masterful v{version} is loaded."
  )


def _get_nagware_message(current_date: datetime,
                         masterful_home_creation_date: datetime,
                         masterful_version: str) -> str:
  """Gets a nagware message given date constraints.

  Returns a nagware message given the current date and the date where the
  ~/.masterful folder was created.

  Args:
    current_date: The current time in datetime format.
    masterful_home_creation_date: The date in which the masterful folder was created.
    masterful_version: The package version to include in the message.

  Returns:
    The message to be printed into the console.
  """
  delta = current_date - masterful_home_creation_date
  message = None
  if delta.days <= _MASTERFUL_REMINDER_DAYS:
    days_to_sign = _MASTERFUL_TRIAL_DAYS - delta.days
    message = _MASTERFUL_REMINDER_DAYS_MESSAGE.format(masterful_version, days_to_sign)

  elif delta.days > _MASTERFUL_REMINDER_DAYS and delta.days <= _MASTERFUL_TRIAL_DAYS:
    message = _MASTERFUL_REMINDER_SIGN_UP_NOW.format(masterful_version)

  elif delta.days > _MASTERFUL_TRIAL_DAYS:
    message = _MASTERFUL_REMINDER_TRIAL_VERSION_FINISHED.format(masterful_version)

  return message


def _print_nagware_reminder(masterful_version: str) -> None:
  """Prints to the console a nagware message.

  Prints Messages to the users indicating when they need to sign up in masterful's
  webpage in order to get a key.

  Args:
    version: The package version to include in the message.
  """
  masterful_home = f'{os.path.expanduser("~")}/.masterful'
  masterful_stats_file = '.stats'
  stats_file_path = os.path.join(masterful_home, masterful_stats_file)

  if os.path.exists(stats_file_path):
    with open(stats_file_path, 'r') as f:
      masterful_home_creation_timestamp = f.read()

    masterful_home_creation_date = datetime.fromtimestamp(float(masterful_home_creation_timestamp))

  else:
    if not os.path.exists(masterful_home):
      os.mkdir(masterful_home)

    with open(stats_file_path, 'w') as f:
      masterful_home_creation_date = datetime.utcnow()
      f.write(str(datetime.timestamp(masterful_home_creation_date)))

  current_date = datetime.utcnow()
  delta = current_date - masterful_home_creation_date
  nagware_message = _get_nagware_message(current_date, masterful_home_creation_date, masterful_version)

  print(nagware_message)
  delta = current_date - masterful_home_creation_date
  if delta.days > _MASTERFUL_REMINDER_DAYS and delta.days <= _MASTERFUL_TRIAL_DAYS:
    print(f"[Pausing for {_NAGWARE_SLEEP_SECONDS} seconds]")
    time.sleep(_NAGWARE_SLEEP_SECONDS)

  if delta.days > _MASTERFUL_TRIAL_DAYS:
    sys.exit('Masterful Library not loaded')



def register(account_id: str = '',
             authorization_key: str = '') -> types.ModuleType:
  """Authorizes the Masterful library and prepares the system for its use.

  This function authorizes usage of the Masterful library, and loads it into
  memory for use. This function *must* be called before using or importing
  any other Masterful packages.

  Note this function returns a new Masterful module instance, so you must
  replace the existing instance like so::

    import masterful
    masterful = masterful.register(...)

  If you need credentials for using Masterful, please request them
  `here <https://www.masterfulai.com/try-free>`_

  Args:
    account_id: The account credentials that were assigned by Masterful.
    authorization_key: The authorization key assigned by Masterful for the provided account.

  Returns:
    Module containing the latest implementation of the Masterful API.
  """

  def remove_file(file_path: str):
    if os.path.exists(file_path):
      os.remove(file_path)

  # Download zip file of Masterful package
  if not account_id:
    account_id = account_id if account_id else constants.MASTERFUL_ACCOUNT_ID
    if account_id == '':
      account_id = _MASTERFUL_TRIAL_ACCOUNT_ID

  if not authorization_key:
    authorization_key = (authorization_key if authorization_key else
                         constants.MASTERFUL_AUTHORIZATION_KEY)
    if authorization_key == '':
      authorization_key = _MASTERFUL_TRIAL_AUTHORIZATION_KEY

  zip_path = f'{_DOWNLOAD_FOLDER}/masterful-{version.__version__}.zip'
  if not f'{zip_path}/masterful-{version.__version__}/' in sys.path:
    response = requests.get(
        constants.MASTERFUL_LOAD + f'?version={version.__version__}',
        auth=auth.HTTPBasicAuth(account_id, authorization_key),
        stream=True)

    if response.status_code != 200:
      _pretty_error_message(response, account_id)
      sys.exit()

    with open(zip_path, 'wb') as zip_file:
      zip_file.write(response.raw.read())
    atexit.register(lambda: remove_file(zip_path))

    # Prepend zip file to system path
    if zip_path not in sys.path:
      sys.path.insert(0, f'{zip_path}/masterful-{version.__version__}/')

    # Remove masterful from the global modules list
    masterful_module_names = []
    for module_name in sys.modules:
      if 'masterful' in module_name:
        masterful_module_names.append(module_name)

    for module_name in masterful_module_names:
      del sys.modules[module_name]

    # Re-import the new masterful
    globals()['masterful'] = __import__('masterful')

  import masterful

  # Update the constants so the rest of the package
  # can use them
  import masterful.constants
  masterful.constants.MASTERFUL_ACCOUNT_ID = account_id
  masterful.constants.MASTERFUL_AUTHORIZATION_KEY = authorization_key
  masterful.constants.MASTERFUL_SESSION_ID = str(uuid.uuid4())
  # uuid.get_node gets the mac address of the device as a 48-bit positive integer
  masterful.constants.MASTERFUL_USER_MAC_ADDRESS = str(uuid.getnode())

  if account_id != _MASTERFUL_TRIAL_ACCOUNT_ID:
    _print_success_message(version.__version__)

  if account_id == _MASTERFUL_TRIAL_ACCOUNT_ID:
    _print_nagware_reminder(version.__version__)

  import masterful.proto.log_pb2
  import masterful.utils.logging

  masterful.utils.logging.log_api_event2(masterful.proto.log_pb2.ApiEventType.REGISTER)

  return masterful



"""Optimization focused parameters and learning algorithms. """
from dataclasses import dataclass
import tensorflow as tf
from typing import Dict, Optional, Sequence, Union


@dataclass
class OptimizationParams:
  """Parameters controlling the optimization of the model during training.

  Optimization means finding the best weights for a
  model and training data. Optimization is different from
  regularization because optimization does not consider
  generalization to unseen data. The central challenge of
  optimization is speed - find the best weights faster.

  Args:
    batch_size: Batch size of the features and labels to use
      during training.
    drop_remainder: True if any examples that do not fit evenly
      into a full batch should be dropped, False otherwise. If False,
      the last batch during training may or may not equal `batch_size`.
    epochs: The number of epochs to train for. An epoch is defined
      as one full sweep through the training data.
    learning_rate: The initial learning rate to use for training. This
      will also be the starting learning rate post-warmup, if warmup
      is specified.
    learning_rate_schedule: An instance if :class:`tf.keras.optimizers.schedules.LearningRateSchedule`
      which modulate how the learning rate of your optimizer changes over time.
      Only one of `learning_rate_schedule` or `learning_rate_callback`
      must be set.
    learning_rate_callback: An instance of :class:`tf.keras.callbacks.Callback`
      which modulate how the learning rate of your optimizer changes over time.
      Only one of `learning_rate_schedule` or `learning_rate_callback`
      must be set.
    warmup_learning_rate: The initial learning rate to use during warmup.
      If `warmup_epochs` is greater than zero, the learning rate will
      linearly grow from `warmup_learning_rate` to `learning_rate` over
      `warmup_epochs` training epochs, incrementing at each step in
      each epoch.
    warmup_epochs: The number of epochs to perform model warmup during
      training. An epoch is defined as one full sweep through the
      training data.
    optimizer: An instance of :class:`tf.keras.optimizers.Optimizer`,
      to use as the optimizer during training.
    loss: An instance of :class:`tf.keras.losses.Loss`, which is the
      function that will be optmized by the optimizer. If the model
      support multiple outputs, this can be a dictionary of loss functions.
    loss_weights: Optional list or dictionary specifying scalar
      coefficients (Python floats) to weight the loss
      contributions of different model outputs. The loss value
      that will be minimized by the model will then be the
      weighted sum of all individual losses, weighted by the
      loss_weights coefficients. If a list, it is expected to
      have a 1:1 mapping to the model's outputs. If a dict,
      it is expected to map output names (strings) to scalar
      coefficients.
    early_stopping_callback: An instance of :class:`tf.keras.callbacks.Callback`
      which can stop training before `total_steps` have been reached depending
      on certain conditions.
    metrics: An optional list of additional metrics that will be measured
      and reported during training.
    readonly_callbacks: An optional list of callbacks to pass to the
      training loop. These callbacks must be readonly. They must not
      attempt to modify any parameters of the training loop, other
      than entries in the logs.
  """
  batch_size: int = None
  drop_remainder: bool = False
  epochs: int = None

  learning_rate: float = None
  learning_rate_schedule: Optional[
      tf.keras.optimizers.schedules.LearningRateSchedule] = None
  learning_rate_callback: Optional[tf.keras.callbacks.Callback] = None

  warmup_learning_rate: float = None
  warmup_epochs: int = None

  optimizer: tf.keras.optimizers.Optimizer = None
  loss: Union[tf.keras.losses.Loss, Dict[str, tf.keras.losses.Loss]] = None
  loss_weights: Union[Sequence, Dict] = None
  early_stopping_callback: Optional[tf.keras.callbacks.Callback] = None

  metrics: Optional[Sequence[tf.keras.metrics.Metric]] = None
  readonly_callbacks: Optional[Sequence[tf.keras.callbacks.Callback]] = None


def learn_optimization_params(
    model: tf.keras.Model,
    model_params: "masterful.architecture.ArchitectureParams",
    dataset: tf.data.Dataset,
    dataset_params: "masterful.data.DataParams",
    **kwargs,
) -> OptimizationParams:
  """Learns the optimal set of optimization parameters for training.

  Example:

  .. code-block:: python

    model: tf.keras.Model = ...
    model_params: masterful.architecture.params.ArchitectureParams = ...
    training_dataset: tf.data.Dataset = ...
    training_dataset_params: masterful.data.params.DataParams = ...
    optimization_params = masterful.optimization.learn_optimization_params(
        model, model_params, training_dataset, training_dataset_params)

  Args:
    model: The model to learn the optimal set of parameters for.
    model_params: The parameters of the given model.
    dataset: The dataset to learn the optimization parameters for.
    dataset_params: Parameters of the dataset used.

  Returns:
    An instance of :class:`OptimizationParams` that can be used during
    training.
  """
  raise RuntimeError(
      "Please call masterful.register() with your account ID and authorization key before using the API."
  )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: masterful/proto/log.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='masterful/proto/log.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x19masterful/proto/log.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"y\n\x08\x41piEvent\x12\x1b\n\x04type\x18\x01 \x01(\x0e\x32\r.ApiEventType\x12\x12\n\nmodel_spec\x18\x02 \x01(\x0c\x12\x18\n\x10\x65xtra_model_spec\x18\x03 \x01(\x0c\x12\x11\n\tdata_spec\x18\x04 \x01(\x0c\x12\x0f\n\x07version\x18\x05 \x01(\x05\"\xb7\x01\n\x0c\x44\x61tasetEvent\x12\x1b\n\x13labeled_cardinality\x18\x01 \x01(\x05\x12\x1e\n\x16validation_cardinality\x18\x02 \x01(\x05\x12\x1d\n\x15unlabeled_cardinality\x18\x03 \x01(\x05\x12\x1d\n\x15synthetic_cardinality\x18\x04 \x01(\x05\x12,\n\x15source_api_event_type\x18\x05 \x01(\x0e\x32\r.ApiEventType\"F\n\x08GuiEvent\x12\x1b\n\x04type\x18\x01 \x01(\x0e\x32\r.GuiEventType\x12\x1d\n\x15policy_search_task_pb\x18\x02 \x01(\x0c\"\xb6\x03\n\x11SearchOnNodeEvent\x12\x37\n\x0bval_results\x18\x01 \x03(\x0b\x32\".SearchOnNodeEvent.ValResultsEntry\x12\x39\n\x0ctest_results\x18\x02 \x03(\x0b\x32#.SearchOnNodeEvent.TestResultsEntry\x1a|\n\x10SearchOnNodeData\x12;\n\x04\x64\x61ta\x18\x01 \x03(\x0b\x32-.SearchOnNodeEvent.SearchOnNodeData.DataEntry\x1a+\n\tDataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\x1aV\n\x0fValResultsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32#.SearchOnNodeEvent.SearchOnNodeData:\x02\x38\x01\x1aW\n\x10TestResultsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32#.SearchOnNodeEvent.SearchOnNodeData:\x02\x38\x01\"\xca\x03\n\x10\x46itCompleteEvent\x12\x1c\n\x14\x65lapsed_time_seconds\x18\x01 \x01(\x02\x12\x36\n\x0bval_results\x18\x02 \x03(\x0b\x32!.FitCompleteEvent.ValResultsEntry\x12\x38\n\x0ctest_results\x18\x03 \x03(\x0b\x32\".FitCompleteEvent.TestResultsEntry\x1ay\n\x0f\x46itCompleteData\x12\x39\n\x04\x64\x61ta\x18\x01 \x03(\x0b\x32+.FitCompleteEvent.FitCompleteData.DataEntry\x1a+\n\tDataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\x1aT\n\x0fValResultsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x30\n\x05value\x18\x02 \x01(\x0b\x32!.FitCompleteEvent.FitCompleteData:\x02\x38\x01\x1aU\n\x10TestResultsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x30\n\x05value\x18\x02 \x01(\x0b\x32!.FitCompleteEvent.FitCompleteData:\x02\x38\x01\"\xf3\x03\n\x14\x41nalyzeCompleteEvent\x12\x1c\n\x14\x65lapsed_time_seconds\x18\x01 \x01(\x02\x12:\n\x0bval_results\x18\x02 \x03(\x0b\x32%.AnalyzeCompleteEvent.ValResultsEntry\x12<\n\x0ctest_results\x18\x03 \x03(\x0b\x32&.AnalyzeCompleteEvent.TestResultsEntry\x1a\x85\x01\n\x13\x41nalyzeCompleteData\x12\x41\n\x04\x64\x61ta\x18\x01 \x03(\x0b\x32\x33.AnalyzeCompleteEvent.AnalyzeCompleteData.DataEntry\x1a+\n\tDataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\x1a\\\n\x0fValResultsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x38\n\x05value\x18\x02 \x01(\x0b\x32).AnalyzeCompleteEvent.AnalyzeCompleteData:\x02\x38\x01\x1a]\n\x10TestResultsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x38\n\x05value\x18\x02 \x01(\x0b\x32).AnalyzeCompleteEvent.AnalyzeCompleteData:\x02\x38\x01\"?\n\x0bPolicyEvent\x12 \n\x0bpolicy_type\x18\x01 \x01(\x0e\x32\x0b.PolicyType\x12\x0e\n\x06policy\x18\x02 \x01(\x0c\"S\n\rApiErrorEvent\x12\x12\n\nerror_type\x18\x01 \x01(\t\x12\x15\n\rerror_context\x18\x02 \x01(\t\x12\x17\n\x0f\x65rror_traceback\x18\x03 \x01(\t\"\xbc\x04\n\x08LogEntry\x12\x13\n\x0b\x63ustomer_id\x18\x01 \x01(\t\x12\x12\n\nsession_id\x18\x02 \x01(\t\x12\x16\n\x0e\x63lient_version\x18\x03 \x01(\t\x12-\n\ttimestamp\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x1e\n\nevent_type\x18\x05 \x01(\x0e\x32\n.EventType\x12#\n\x0e\x61pi_event_data\x18\x06 \x01(\x0b\x32\t.ApiEventH\x00\x12+\n\x12\x64\x61taset_event_data\x18\x07 \x01(\x0b\x32\r.DatasetEventH\x00\x12\x37\n\x19search_on_node_event_data\x18\x08 \x01(\x0b\x32\x12.SearchOnNodeEventH\x00\x12\x34\n\x17\x66it_complete_event_data\x18\t \x01(\x0b\x32\x11.FitCompleteEventH\x00\x12<\n\x1b\x61nalyze_complete_event_data\x18\n \x01(\x0b\x32\x15.AnalyzeCompleteEventH\x00\x12#\n\x0egui_event_data\x18\x0b \x01(\x0b\x32\t.GuiEventH\x00\x12)\n\x11policy_event_data\x18\x0c \x01(\x0b\x32\x0c.PolicyEventH\x00\x12.\n\x14\x61pi_error_event_data\x18\r \x01(\x0b\x32\x0e.ApiErrorEventH\x00\x12\x13\n\x0bmac_address\x18\x0e \x01(\tB\x0c\n\nevent_data*\xa6\x02\n\x0c\x41piEventType\x12\x1a\n\x16\x41PI_EVENT_TYPE_UNKNOWN\x10\x00\x12\x0b\n\x07\x41UTOFIT\x10\x01\x12\x18\n\x14\x43ORE_FIND_FIT_POLICY\x10\x02\x12\x0c\n\x08\x43ORE_FIT\x10\x03\x12\x11\n\rCORE_ENSEMBLE\x10\x04\x12\x10\n\x0c\x43ORE_DISTILL\x10\x05\x12\x11\n\rCORE_PRETRAIN\x10\x06\x12!\n\x1d\x43ORE_FIND_AUGMENTATION_POLICY\x10\x07\x12\x12\n\x0eSEARCH_ON_NODE\x10\x08\x12\t\n\x05TRAIN\x10\t\x12\x14\n\x10SSL_ANALYZE_DATA\x10\n\x12\x11\n\rSSL_LOAD_FROM\x10\x0b\x12\x0c\n\x08REGISTER\x10\x0c\x12\x14\n\x10IMPORT_MASTERFUL\x10\r*P\n\x0cGuiEventType\x12\x1a\n\x16GUI_EVENT_TYPE_UNKNOWN\x10\x00\x12$\n SERIALIZED_POLICY_SEARCH_TASK_PB\x10\x01*\x87\x01\n\nPolicyType\x12\x17\n\x13POLICY_TYPE_UNKNOWN\x10\x00\x12\x13\n\x0fPOLICY_TYPE_FIT\x10\x01\x12\x18\n\x14POLICY_TYPE_ENSEMBLE\x10\x02\x12\x17\n\x13POLICY_TYPE_DISTILL\x10\x03\x12\x18\n\x14POLICY_TYPE_PRETRAIN\x10\x04*\xc9\x01\n\tEventType\x12\x16\n\x12\x45VENT_TYPE_UNKNOWN\x10\x00\x12\r\n\tAPI_EVENT\x10\x01\x12\x11\n\rDATASET_EVENT\x10\x02\x12\x18\n\x14SEARCH_ON_NODE_EVENT\x10\x03\x12\x16\n\x12\x46IT_COMPLETE_EVENT\x10\x04\x12\x1a\n\x16\x41NALYZE_COMPLETE_EVENT\x10\x05\x12\r\n\tGUI_EVENT\x10\x06\x12\x10\n\x0cPOLICY_EVENT\x10\x07\x12\x13\n\x0f\x41PI_ERROR_EVENT\x10\x08\x62\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,])

_APIEVENTTYPE = _descriptor.EnumDescriptor(
  name='ApiEventType',
  full_name='ApiEventType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='API_EVENT_TYPE_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AUTOFIT', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CORE_FIND_FIT_POLICY', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CORE_FIT', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CORE_ENSEMBLE', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CORE_DISTILL', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CORE_PRETRAIN', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CORE_FIND_AUGMENTATION_POLICY', index=7, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SEARCH_ON_NODE', index=8, number=8,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TRAIN', index=9, number=9,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SSL_ANALYZE_DATA', index=10, number=10,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SSL_LOAD_FROM', index=11, number=11,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REGISTER', index=12, number=12,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IMPORT_MASTERFUL', index=13, number=13,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2573,
  serialized_end=2867,
)
_sym_db.RegisterEnumDescriptor(_APIEVENTTYPE)

ApiEventType = enum_type_wrapper.EnumTypeWrapper(_APIEVENTTYPE)
_GUIEVENTTYPE = _descriptor.EnumDescriptor(
  name='GuiEventType',
  full_name='GuiEventType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='GUI_EVENT_TYPE_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SERIALIZED_POLICY_SEARCH_TASK_PB', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2869,
  serialized_end=2949,
)
_sym_db.RegisterEnumDescriptor(_GUIEVENTTYPE)

GuiEventType = enum_type_wrapper.EnumTypeWrapper(_GUIEVENTTYPE)
_POLICYTYPE = _descriptor.EnumDescriptor(
  name='PolicyType',
  full_name='PolicyType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='POLICY_TYPE_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='POLICY_TYPE_FIT', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='POLICY_TYPE_ENSEMBLE', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='POLICY_TYPE_DISTILL', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='POLICY_TYPE_PRETRAIN', index=4, number=4,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2952,
  serialized_end=3087,
)
_sym_db.RegisterEnumDescriptor(_POLICYTYPE)

PolicyType = enum_type_wrapper.EnumTypeWrapper(_POLICYTYPE)
_EVENTTYPE = _descriptor.EnumDescriptor(
  name='EventType',
  full_name='EventType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='EVENT_TYPE_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='API_EVENT', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATASET_EVENT', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SEARCH_ON_NODE_EVENT', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FIT_COMPLETE_EVENT', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ANALYZE_COMPLETE_EVENT', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GUI_EVENT', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='POLICY_EVENT', index=7, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='API_ERROR_EVENT', index=8, number=8,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=3090,
  serialized_end=3291,
)
_sym_db.RegisterEnumDescriptor(_EVENTTYPE)

EventType = enum_type_wrapper.EnumTypeWrapper(_EVENTTYPE)
API_EVENT_TYPE_UNKNOWN = 0
AUTOFIT = 1
CORE_FIND_FIT_POLICY = 2
CORE_FIT = 3
CORE_ENSEMBLE = 4
CORE_DISTILL = 5
CORE_PRETRAIN = 6
CORE_FIND_AUGMENTATION_POLICY = 7
SEARCH_ON_NODE = 8
TRAIN = 9
SSL_ANALYZE_DATA = 10
SSL_LOAD_FROM = 11
REGISTER = 12
IMPORT_MASTERFUL = 13
GUI_EVENT_TYPE_UNKNOWN = 0
SERIALIZED_POLICY_SEARCH_TASK_PB = 1
POLICY_TYPE_UNKNOWN = 0
POLICY_TYPE_FIT = 1
POLICY_TYPE_ENSEMBLE = 2
POLICY_TYPE_DISTILL = 3
POLICY_TYPE_PRETRAIN = 4
EVENT_TYPE_UNKNOWN = 0
API_EVENT = 1
DATASET_EVENT = 2
SEARCH_ON_NODE_EVENT = 3
FIT_COMPLETE_EVENT = 4
ANALYZE_COMPLETE_EVENT = 5
GUI_EVENT = 6
POLICY_EVENT = 7
API_ERROR_EVENT = 8



_APIEVENT = _descriptor.Descriptor(
  name='ApiEvent',
  full_name='ApiEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='ApiEvent.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='model_spec', full_name='ApiEvent.model_spec', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='extra_model_spec', full_name='ApiEvent.extra_model_spec', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data_spec', full_name='ApiEvent.data_spec', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version', full_name='ApiEvent.version', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=62,
  serialized_end=183,
)


_DATASETEVENT = _descriptor.Descriptor(
  name='DatasetEvent',
  full_name='DatasetEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='labeled_cardinality', full_name='DatasetEvent.labeled_cardinality', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='validation_cardinality', full_name='DatasetEvent.validation_cardinality', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='unlabeled_cardinality', full_name='DatasetEvent.unlabeled_cardinality', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='synthetic_cardinality', full_name='DatasetEvent.synthetic_cardinality', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='source_api_event_type', full_name='DatasetEvent.source_api_event_type', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=186,
  serialized_end=369,
)


_GUIEVENT = _descriptor.Descriptor(
  name='GuiEvent',
  full_name='GuiEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='GuiEvent.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='policy_search_task_pb', full_name='GuiEvent.policy_search_task_pb', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=371,
  serialized_end=441,
)


_SEARCHONNODEEVENT_SEARCHONNODEDATA_DATAENTRY = _descriptor.Descriptor(
  name='DataEntry',
  full_name='SearchOnNodeEvent.SearchOnNodeData.DataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='SearchOnNodeEvent.SearchOnNodeData.DataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='SearchOnNodeEvent.SearchOnNodeData.DataEntry.value', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=662,
  serialized_end=705,
)

_SEARCHONNODEEVENT_SEARCHONNODEDATA = _descriptor.Descriptor(
  name='SearchOnNodeData',
  full_name='SearchOnNodeEvent.SearchOnNodeData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='SearchOnNodeEvent.SearchOnNodeData.data', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_SEARCHONNODEEVENT_SEARCHONNODEDATA_DATAENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=581,
  serialized_end=705,
)

_SEARCHONNODEEVENT_VALRESULTSENTRY = _descriptor.Descriptor(
  name='ValResultsEntry',
  full_name='SearchOnNodeEvent.ValResultsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='SearchOnNodeEvent.ValResultsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='SearchOnNodeEvent.ValResultsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=707,
  serialized_end=793,
)

_SEARCHONNODEEVENT_TESTRESULTSENTRY = _descriptor.Descriptor(
  name='TestResultsEntry',
  full_name='SearchOnNodeEvent.TestResultsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='SearchOnNodeEvent.TestResultsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='SearchOnNodeEvent.TestResultsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=795,
  serialized_end=882,
)

_SEARCHONNODEEVENT = _descriptor.Descriptor(
  name='SearchOnNodeEvent',
  full_name='SearchOnNodeEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='val_results', full_name='SearchOnNodeEvent.val_results', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='test_results', full_name='SearchOnNodeEvent.test_results', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_SEARCHONNODEEVENT_SEARCHONNODEDATA, _SEARCHONNODEEVENT_VALRESULTSENTRY, _SEARCHONNODEEVENT_TESTRESULTSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=444,
  serialized_end=882,
)


_FITCOMPLETEEVENT_FITCOMPLETEDATA_DATAENTRY = _descriptor.Descriptor(
  name='DataEntry',
  full_name='FitCompleteEvent.FitCompleteData.DataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='FitCompleteEvent.FitCompleteData.DataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='FitCompleteEvent.FitCompleteData.DataEntry.value', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=662,
  serialized_end=705,
)

_FITCOMPLETEEVENT_FITCOMPLETEDATA = _descriptor.Descriptor(
  name='FitCompleteData',
  full_name='FitCompleteEvent.FitCompleteData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='FitCompleteEvent.FitCompleteData.data', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_FITCOMPLETEEVENT_FITCOMPLETEDATA_DATAENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1049,
  serialized_end=1170,
)

_FITCOMPLETEEVENT_VALRESULTSENTRY = _descriptor.Descriptor(
  name='ValResultsEntry',
  full_name='FitCompleteEvent.ValResultsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='FitCompleteEvent.ValResultsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='FitCompleteEvent.ValResultsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1172,
  serialized_end=1256,
)

_FITCOMPLETEEVENT_TESTRESULTSENTRY = _descriptor.Descriptor(
  name='TestResultsEntry',
  full_name='FitCompleteEvent.TestResultsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='FitCompleteEvent.TestResultsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='FitCompleteEvent.TestResultsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1258,
  serialized_end=1343,
)

_FITCOMPLETEEVENT = _descriptor.Descriptor(
  name='FitCompleteEvent',
  full_name='FitCompleteEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='elapsed_time_seconds', full_name='FitCompleteEvent.elapsed_time_seconds', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='val_results', full_name='FitCompleteEvent.val_results', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='test_results', full_name='FitCompleteEvent.test_results', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_FITCOMPLETEEVENT_FITCOMPLETEDATA, _FITCOMPLETEEVENT_VALRESULTSENTRY, _FITCOMPLETEEVENT_TESTRESULTSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=885,
  serialized_end=1343,
)


_ANALYZECOMPLETEEVENT_ANALYZECOMPLETEDATA_DATAENTRY = _descriptor.Descriptor(
  name='DataEntry',
  full_name='AnalyzeCompleteEvent.AnalyzeCompleteData.DataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='AnalyzeCompleteEvent.AnalyzeCompleteData.DataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='AnalyzeCompleteEvent.AnalyzeCompleteData.DataEntry.value', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=662,
  serialized_end=705,
)

_ANALYZECOMPLETEEVENT_ANALYZECOMPLETEDATA = _descriptor.Descriptor(
  name='AnalyzeCompleteData',
  full_name='AnalyzeCompleteEvent.AnalyzeCompleteData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='AnalyzeCompleteEvent.AnalyzeCompleteData.data', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_ANALYZECOMPLETEEVENT_ANALYZECOMPLETEDATA_DATAENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1523,
  serialized_end=1656,
)

_ANALYZECOMPLETEEVENT_VALRESULTSENTRY = _descriptor.Descriptor(
  name='ValResultsEntry',
  full_name='AnalyzeCompleteEvent.ValResultsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='AnalyzeCompleteEvent.ValResultsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='AnalyzeCompleteEvent.ValResultsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1658,
  serialized_end=1750,
)

_ANALYZECOMPLETEEVENT_TESTRESULTSENTRY = _descriptor.Descriptor(
  name='TestResultsEntry',
  full_name='AnalyzeCompleteEvent.TestResultsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='AnalyzeCompleteEvent.TestResultsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='AnalyzeCompleteEvent.TestResultsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1752,
  serialized_end=1845,
)

_ANALYZECOMPLETEEVENT = _descriptor.Descriptor(
  name='AnalyzeCompleteEvent',
  full_name='AnalyzeCompleteEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='elapsed_time_seconds', full_name='AnalyzeCompleteEvent.elapsed_time_seconds', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='val_results', full_name='AnalyzeCompleteEvent.val_results', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='test_results', full_name='AnalyzeCompleteEvent.test_results', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_ANALYZECOMPLETEEVENT_ANALYZECOMPLETEDATA, _ANALYZECOMPLETEEVENT_VALRESULTSENTRY, _ANALYZECOMPLETEEVENT_TESTRESULTSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1346,
  serialized_end=1845,
)


_POLICYEVENT = _descriptor.Descriptor(
  name='PolicyEvent',
  full_name='PolicyEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='policy_type', full_name='PolicyEvent.policy_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='policy', full_name='PolicyEvent.policy', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1847,
  serialized_end=1910,
)


_APIERROREVENT = _descriptor.Descriptor(
  name='ApiErrorEvent',
  full_name='ApiErrorEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='error_type', full_name='ApiErrorEvent.error_type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error_context', full_name='ApiErrorEvent.error_context', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error_traceback', full_name='ApiErrorEvent.error_traceback', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1912,
  serialized_end=1995,
)


_LOGENTRY = _descriptor.Descriptor(
  name='LogEntry',
  full_name='LogEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='customer_id', full_name='LogEntry.customer_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='session_id', full_name='LogEntry.session_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='client_version', full_name='LogEntry.client_version', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='LogEntry.timestamp', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='event_type', full_name='LogEntry.event_type', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='api_event_data', full_name='LogEntry.api_event_data', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dataset_event_data', full_name='LogEntry.dataset_event_data', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='search_on_node_event_data', full_name='LogEntry.search_on_node_event_data', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fit_complete_event_data', full_name='LogEntry.fit_complete_event_data', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='analyze_complete_event_data', full_name='LogEntry.analyze_complete_event_data', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='gui_event_data', full_name='LogEntry.gui_event_data', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='policy_event_data', full_name='LogEntry.policy_event_data', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='api_error_event_data', full_name='LogEntry.api_error_event_data', index=12,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mac_address', full_name='LogEntry.mac_address', index=13,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='event_data', full_name='LogEntry.event_data',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=1998,
  serialized_end=2570,
)

_APIEVENT.fields_by_name['type'].enum_type = _APIEVENTTYPE
_DATASETEVENT.fields_by_name['source_api_event_type'].enum_type = _APIEVENTTYPE
_GUIEVENT.fields_by_name['type'].enum_type = _GUIEVENTTYPE
_SEARCHONNODEEVENT_SEARCHONNODEDATA_DATAENTRY.containing_type = _SEARCHONNODEEVENT_SEARCHONNODEDATA
_SEARCHONNODEEVENT_SEARCHONNODEDATA.fields_by_name['data'].message_type = _SEARCHONNODEEVENT_SEARCHONNODEDATA_DATAENTRY
_SEARCHONNODEEVENT_SEARCHONNODEDATA.containing_type = _SEARCHONNODEEVENT
_SEARCHONNODEEVENT_VALRESULTSENTRY.fields_by_name['value'].message_type = _SEARCHONNODEEVENT_SEARCHONNODEDATA
_SEARCHONNODEEVENT_VALRESULTSENTRY.containing_type = _SEARCHONNODEEVENT
_SEARCHONNODEEVENT_TESTRESULTSENTRY.fields_by_name['value'].message_type = _SEARCHONNODEEVENT_SEARCHONNODEDATA
_SEARCHONNODEEVENT_TESTRESULTSENTRY.containing_type = _SEARCHONNODEEVENT
_SEARCHONNODEEVENT.fields_by_name['val_results'].message_type = _SEARCHONNODEEVENT_VALRESULTSENTRY
_SEARCHONNODEEVENT.fields_by_name['test_results'].message_type = _SEARCHONNODEEVENT_TESTRESULTSENTRY
_FITCOMPLETEEVENT_FITCOMPLETEDATA_DATAENTRY.containing_type = _FITCOMPLETEEVENT_FITCOMPLETEDATA
_FITCOMPLETEEVENT_FITCOMPLETEDATA.fields_by_name['data'].message_type = _FITCOMPLETEEVENT_FITCOMPLETEDATA_DATAENTRY
_FITCOMPLETEEVENT_FITCOMPLETEDATA.containing_type = _FITCOMPLETEEVENT
_FITCOMPLETEEVENT_VALRESULTSENTRY.fields_by_name['value'].message_type = _FITCOMPLETEEVENT_FITCOMPLETEDATA
_FITCOMPLETEEVENT_VALRESULTSENTRY.containing_type = _FITCOMPLETEEVENT
_FITCOMPLETEEVENT_TESTRESULTSENTRY.fields_by_name['value'].message_type = _FITCOMPLETEEVENT_FITCOMPLETEDATA
_FITCOMPLETEEVENT_TESTRESULTSENTRY.containing_type = _FITCOMPLETEEVENT
_FITCOMPLETEEVENT.fields_by_name['val_results'].message_type = _FITCOMPLETEEVENT_VALRESULTSENTRY
_FITCOMPLETEEVENT.fields_by_name['test_results'].message_type = _FITCOMPLETEEVENT_TESTRESULTSENTRY
_ANALYZECOMPLETEEVENT_ANALYZECOMPLETEDATA_DATAENTRY.containing_type = _ANALYZECOMPLETEEVENT_ANALYZECOMPLETEDATA
_ANALYZECOMPLETEEVENT_ANALYZECOMPLETEDATA.fields_by_name['data'].message_type = _ANALYZECOMPLETEEVENT_ANALYZECOMPLETEDATA_DATAENTRY
_ANALYZECOMPLETEEVENT_ANALYZECOMPLETEDATA.containing_type = _ANALYZECOMPLETEEVENT
_ANALYZECOMPLETEEVENT_VALRESULTSENTRY.fields_by_name['value'].message_type = _ANALYZECOMPLETEEVENT_ANALYZECOMPLETEDATA
_ANALYZECOMPLETEEVENT_VALRESULTSENTRY.containing_type = _ANALYZECOMPLETEEVENT
_ANALYZECOMPLETEEVENT_TESTRESULTSENTRY.fields_by_name['value'].message_type = _ANALYZECOMPLETEEVENT_ANALYZECOMPLETEDATA
_ANALYZECOMPLETEEVENT_TESTRESULTSENTRY.containing_type = _ANALYZECOMPLETEEVENT
_ANALYZECOMPLETEEVENT.fields_by_name['val_results'].message_type = _ANALYZECOMPLETEEVENT_VALRESULTSENTRY
_ANALYZECOMPLETEEVENT.fields_by_name['test_results'].message_type = _ANALYZECOMPLETEEVENT_TESTRESULTSENTRY
_POLICYEVENT.fields_by_name['policy_type'].enum_type = _POLICYTYPE
_LOGENTRY.fields_by_name['timestamp'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LOGENTRY.fields_by_name['event_type'].enum_type = _EVENTTYPE
_LOGENTRY.fields_by_name['api_event_data'].message_type = _APIEVENT
_LOGENTRY.fields_by_name['dataset_event_data'].message_type = _DATASETEVENT
_LOGENTRY.fields_by_name['search_on_node_event_data'].message_type = _SEARCHONNODEEVENT
_LOGENTRY.fields_by_name['fit_complete_event_data'].message_type = _FITCOMPLETEEVENT
_LOGENTRY.fields_by_name['analyze_complete_event_data'].message_type = _ANALYZECOMPLETEEVENT
_LOGENTRY.fields_by_name['gui_event_data'].message_type = _GUIEVENT
_LOGENTRY.fields_by_name['policy_event_data'].message_type = _POLICYEVENT
_LOGENTRY.fields_by_name['api_error_event_data'].message_type = _APIERROREVENT
_LOGENTRY.oneofs_by_name['event_data'].fields.append(
  _LOGENTRY.fields_by_name['api_event_data'])
_LOGENTRY.fields_by_name['api_event_data'].containing_oneof = _LOGENTRY.oneofs_by_name['event_data']
_LOGENTRY.oneofs_by_name['event_data'].fields.append(
  _LOGENTRY.fields_by_name['dataset_event_data'])
_LOGENTRY.fields_by_name['dataset_event_data'].containing_oneof = _LOGENTRY.oneofs_by_name['event_data']
_LOGENTRY.oneofs_by_name['event_data'].fields.append(
  _LOGENTRY.fields_by_name['search_on_node_event_data'])
_LOGENTRY.fields_by_name['search_on_node_event_data'].containing_oneof = _LOGENTRY.oneofs_by_name['event_data']
_LOGENTRY.oneofs_by_name['event_data'].fields.append(
  _LOGENTRY.fields_by_name['fit_complete_event_data'])
_LOGENTRY.fields_by_name['fit_complete_event_data'].containing_oneof = _LOGENTRY.oneofs_by_name['event_data']
_LOGENTRY.oneofs_by_name['event_data'].fields.append(
  _LOGENTRY.fields_by_name['analyze_complete_event_data'])
_LOGENTRY.fields_by_name['analyze_complete_event_data'].containing_oneof = _LOGENTRY.oneofs_by_name['event_data']
_LOGENTRY.oneofs_by_name['event_data'].fields.append(
  _LOGENTRY.fields_by_name['gui_event_data'])
_LOGENTRY.fields_by_name['gui_event_data'].containing_oneof = _LOGENTRY.oneofs_by_name['event_data']
_LOGENTRY.oneofs_by_name['event_data'].fields.append(
  _LOGENTRY.fields_by_name['policy_event_data'])
_LOGENTRY.fields_by_name['policy_event_data'].containing_oneof = _LOGENTRY.oneofs_by_name['event_data']
_LOGENTRY.oneofs_by_name['event_data'].fields.append(
  _LOGENTRY.fields_by_name['api_error_event_data'])
_LOGENTRY.fields_by_name['api_error_event_data'].containing_oneof = _LOGENTRY.oneofs_by_name['event_data']
DESCRIPTOR.message_types_by_name['ApiEvent'] = _APIEVENT
DESCRIPTOR.message_types_by_name['DatasetEvent'] = _DATASETEVENT
DESCRIPTOR.message_types_by_name['GuiEvent'] = _GUIEVENT
DESCRIPTOR.message_types_by_name['SearchOnNodeEvent'] = _SEARCHONNODEEVENT
DESCRIPTOR.message_types_by_name['FitCompleteEvent'] = _FITCOMPLETEEVENT
DESCRIPTOR.message_types_by_name['AnalyzeCompleteEvent'] = _ANALYZECOMPLETEEVENT
DESCRIPTOR.message_types_by_name['PolicyEvent'] = _POLICYEVENT
DESCRIPTOR.message_types_by_name['ApiErrorEvent'] = _APIERROREVENT
DESCRIPTOR.message_types_by_name['LogEntry'] = _LOGENTRY
DESCRIPTOR.enum_types_by_name['ApiEventType'] = _APIEVENTTYPE
DESCRIPTOR.enum_types_by_name['GuiEventType'] = _GUIEVENTTYPE
DESCRIPTOR.enum_types_by_name['PolicyType'] = _POLICYTYPE
DESCRIPTOR.enum_types_by_name['EventType'] = _EVENTTYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ApiEvent = _reflection.GeneratedProtocolMessageType('ApiEvent', (_message.Message,), {
  'DESCRIPTOR' : _APIEVENT,
  '__module__' : 'masterful.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:ApiEvent)
  })
_sym_db.RegisterMessage(ApiEvent)

DatasetEvent = _reflection.GeneratedProtocolMessageType('DatasetEvent', (_message.Message,), {
  'DESCRIPTOR' : _DATASETEVENT,
  '__module__' : 'masterful.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:DatasetEvent)
  })
_sym_db.RegisterMessage(DatasetEvent)

GuiEvent = _reflection.GeneratedProtocolMessageType('GuiEvent', (_message.Message,), {
  'DESCRIPTOR' : _GUIEVENT,
  '__module__' : 'masterful.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:GuiEvent)
  })
_sym_db.RegisterMessage(GuiEvent)

SearchOnNodeEvent = _reflection.GeneratedProtocolMessageType('SearchOnNodeEvent', (_message.Message,), {

  'SearchOnNodeData' : _reflection.GeneratedProtocolMessageType('SearchOnNodeData', (_message.Message,), {

    'DataEntry' : _reflection.GeneratedProtocolMessageType('DataEntry', (_message.Message,), {
      'DESCRIPTOR' : _SEARCHONNODEEVENT_SEARCHONNODEDATA_DATAENTRY,
      '__module__' : 'masterful.proto.log_pb2'
      # @@protoc_insertion_point(class_scope:SearchOnNodeEvent.SearchOnNodeData.DataEntry)
      })
    ,
    'DESCRIPTOR' : _SEARCHONNODEEVENT_SEARCHONNODEDATA,
    '__module__' : 'masterful.proto.log_pb2'
    # @@protoc_insertion_point(class_scope:SearchOnNodeEvent.SearchOnNodeData)
    })
  ,

  'ValResultsEntry' : _reflection.GeneratedProtocolMessageType('ValResultsEntry', (_message.Message,), {
    'DESCRIPTOR' : _SEARCHONNODEEVENT_VALRESULTSENTRY,
    '__module__' : 'masterful.proto.log_pb2'
    # @@protoc_insertion_point(class_scope:SearchOnNodeEvent.ValResultsEntry)
    })
  ,

  'TestResultsEntry' : _reflection.GeneratedProtocolMessageType('TestResultsEntry', (_message.Message,), {
    'DESCRIPTOR' : _SEARCHONNODEEVENT_TESTRESULTSENTRY,
    '__module__' : 'masterful.proto.log_pb2'
    # @@protoc_insertion_point(class_scope:SearchOnNodeEvent.TestResultsEntry)
    })
  ,
  'DESCRIPTOR' : _SEARCHONNODEEVENT,
  '__module__' : 'masterful.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:SearchOnNodeEvent)
  })
_sym_db.RegisterMessage(SearchOnNodeEvent)
_sym_db.RegisterMessage(SearchOnNodeEvent.SearchOnNodeData)
_sym_db.RegisterMessage(SearchOnNodeEvent.SearchOnNodeData.DataEntry)
_sym_db.RegisterMessage(SearchOnNodeEvent.ValResultsEntry)
_sym_db.RegisterMessage(SearchOnNodeEvent.TestResultsEntry)

FitCompleteEvent = _reflection.GeneratedProtocolMessageType('FitCompleteEvent', (_message.Message,), {

  'FitCompleteData' : _reflection.GeneratedProtocolMessageType('FitCompleteData', (_message.Message,), {

    'DataEntry' : _reflection.GeneratedProtocolMessageType('DataEntry', (_message.Message,), {
      'DESCRIPTOR' : _FITCOMPLETEEVENT_FITCOMPLETEDATA_DATAENTRY,
      '__module__' : 'masterful.proto.log_pb2'
      # @@protoc_insertion_point(class_scope:FitCompleteEvent.FitCompleteData.DataEntry)
      })
    ,
    'DESCRIPTOR' : _FITCOMPLETEEVENT_FITCOMPLETEDATA,
    '__module__' : 'masterful.proto.log_pb2'
    # @@protoc_insertion_point(class_scope:FitCompleteEvent.FitCompleteData)
    })
  ,

  'ValResultsEntry' : _reflection.GeneratedProtocolMessageType('ValResultsEntry', (_message.Message,), {
    'DESCRIPTOR' : _FITCOMPLETEEVENT_VALRESULTSENTRY,
    '__module__' : 'masterful.proto.log_pb2'
    # @@protoc_insertion_point(class_scope:FitCompleteEvent.ValResultsEntry)
    })
  ,

  'TestResultsEntry' : _reflection.GeneratedProtocolMessageType('TestResultsEntry', (_message.Message,), {
    'DESCRIPTOR' : _FITCOMPLETEEVENT_TESTRESULTSENTRY,
    '__module__' : 'masterful.proto.log_pb2'
    # @@protoc_insertion_point(class_scope:FitCompleteEvent.TestResultsEntry)
    })
  ,
  'DESCRIPTOR' : _FITCOMPLETEEVENT,
  '__module__' : 'masterful.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:FitCompleteEvent)
  })
_sym_db.RegisterMessage(FitCompleteEvent)
_sym_db.RegisterMessage(FitCompleteEvent.FitCompleteData)
_sym_db.RegisterMessage(FitCompleteEvent.FitCompleteData.DataEntry)
_sym_db.RegisterMessage(FitCompleteEvent.ValResultsEntry)
_sym_db.RegisterMessage(FitCompleteEvent.TestResultsEntry)

AnalyzeCompleteEvent = _reflection.GeneratedProtocolMessageType('AnalyzeCompleteEvent', (_message.Message,), {

  'AnalyzeCompleteData' : _reflection.GeneratedProtocolMessageType('AnalyzeCompleteData', (_message.Message,), {

    'DataEntry' : _reflection.GeneratedProtocolMessageType('DataEntry', (_message.Message,), {
      'DESCRIPTOR' : _ANALYZECOMPLETEEVENT_ANALYZECOMPLETEDATA_DATAENTRY,
      '__module__' : 'masterful.proto.log_pb2'
      # @@protoc_insertion_point(class_scope:AnalyzeCompleteEvent.AnalyzeCompleteData.DataEntry)
      })
    ,
    'DESCRIPTOR' : _ANALYZECOMPLETEEVENT_ANALYZECOMPLETEDATA,
    '__module__' : 'masterful.proto.log_pb2'
    # @@protoc_insertion_point(class_scope:AnalyzeCompleteEvent.AnalyzeCompleteData)
    })
  ,

  'ValResultsEntry' : _reflection.GeneratedProtocolMessageType('ValResultsEntry', (_message.Message,), {
    'DESCRIPTOR' : _ANALYZECOMPLETEEVENT_VALRESULTSENTRY,
    '__module__' : 'masterful.proto.log_pb2'
    # @@protoc_insertion_point(class_scope:AnalyzeCompleteEvent.ValResultsEntry)
    })
  ,

  'TestResultsEntry' : _reflection.GeneratedProtocolMessageType('TestResultsEntry', (_message.Message,), {
    'DESCRIPTOR' : _ANALYZECOMPLETEEVENT_TESTRESULTSENTRY,
    '__module__' : 'masterful.proto.log_pb2'
    # @@protoc_insertion_point(class_scope:AnalyzeCompleteEvent.TestResultsEntry)
    })
  ,
  'DESCRIPTOR' : _ANALYZECOMPLETEEVENT,
  '__module__' : 'masterful.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:AnalyzeCompleteEvent)
  })
_sym_db.RegisterMessage(AnalyzeCompleteEvent)
_sym_db.RegisterMessage(AnalyzeCompleteEvent.AnalyzeCompleteData)
_sym_db.RegisterMessage(AnalyzeCompleteEvent.AnalyzeCompleteData.DataEntry)
_sym_db.RegisterMessage(AnalyzeCompleteEvent.ValResultsEntry)
_sym_db.RegisterMessage(AnalyzeCompleteEvent.TestResultsEntry)

PolicyEvent = _reflection.GeneratedProtocolMessageType('PolicyEvent', (_message.Message,), {
  'DESCRIPTOR' : _POLICYEVENT,
  '__module__' : 'masterful.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:PolicyEvent)
  })
_sym_db.RegisterMessage(PolicyEvent)

ApiErrorEvent = _reflection.GeneratedProtocolMessageType('ApiErrorEvent', (_message.Message,), {
  'DESCRIPTOR' : _APIERROREVENT,
  '__module__' : 'masterful.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:ApiErrorEvent)
  })
_sym_db.RegisterMessage(ApiErrorEvent)

LogEntry = _reflection.GeneratedProtocolMessageType('LogEntry', (_message.Message,), {
  'DESCRIPTOR' : _LOGENTRY,
  '__module__' : 'masterful.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:LogEntry)
  })
_sym_db.RegisterMessage(LogEntry)


_SEARCHONNODEEVENT_SEARCHONNODEDATA_DATAENTRY._options = None
_SEARCHONNODEEVENT_VALRESULTSENTRY._options = None
_SEARCHONNODEEVENT_TESTRESULTSENTRY._options = None
_FITCOMPLETEEVENT_FITCOMPLETEDATA_DATAENTRY._options = None
_FITCOMPLETEEVENT_VALRESULTSENTRY._options = None
_FITCOMPLETEEVENT_TESTRESULTSENTRY._options = None
_ANALYZECOMPLETEEVENT_ANALYZECOMPLETEDATA_DATAENTRY._options = None
_ANALYZECOMPLETEEVENT_VALRESULTSENTRY._options = None
_ANALYZECOMPLETEEVENT_TESTRESULTSENTRY._options = None
# @@protoc_insertion_point(module_scope)



# Masterful only supports tensorflow 2.4.0

def _check_tensorflow_version():
  from packaging import version
  try:
    import tensorflow as tf
    assert version.parse(tf.__version__) >= version.parse(
        '2.4.0'
    ), "Masterful only supports Tensorflow versions 2.4.0 and greater."
  except ImportError:
    print(
        "Masterful requires Tensorflow 2.4+. Please install Tensorflow before installing Masterful."
    )
    exit()
  

def _log_masterful_import():
  """Logs an import event to datalake."""

  from masterful.proto import log_pb2

  MASTERFUL_STUB_ACCOUNT_ID = 'bb8c34ca-2abb-4e74-83c6-595d8da8b50e'
  MASTERFUL_STUB_AUTHORIZATION_KEY = '3eb905df-69a0-4b34-91e4-56a6dba893ca'

  def _create_base_log_entry():
    import uuid
    from masterful.version import __version__ as masterful_version

    log_entry = log_pb2.LogEntry()
    log_entry.customer_id = MASTERFUL_STUB_ACCOUNT_ID
    log_entry.mac_address = str(uuid.getnode())
    log_entry.client_version = masterful_version
    return log_entry

  def _send_log_entry(log_entry: log_pb2.LogEntry):
    """Sends a log request to the server carrying a log entry payload.

    Request exceptions are intercepted and printed to console.

    Args:
      log_entry: The log entry payload.
    """
    import requests
    from requests import auth
    from masterful import constants

    headers = {'Content-Type': 'application/octet-stream'}
    http_auth = auth.HTTPBasicAuth(MASTERFUL_STUB_ACCOUNT_ID,
                                  MASTERFUL_STUB_AUTHORIZATION_KEY)
    try:
      response = requests.post(constants.MASTERFUL_LOG_ENTRY,
                              headers=headers,
                              data=log_entry.SerializeToString(),
                              auth=http_auth)
      # TODO(MAS-1301): Retrieve global logger for reporting
      if response.status_code not in [200, 201]:
        print(
            f'Log {log_pb2.EventType.keys()[log_pb2.EventType.values().index(log_entry.event_type)]} ({response.status_code}): {response.json()}'
        )
    except requests.exceptions.RequestException as request_exception:
      print(request_exception)

  

  log_entry = _create_base_log_entry()
  log_entry.event_type = log_pb2.EventType.API_EVENT
  log_entry.api_event_data.type = log_pb2.ApiEventType.IMPORT_MASTERFUL
  log_entry.api_event_data.version = 2

  _send_log_entry(log_entry)


_check_tensorflow_version()
_log_masterful_import()

# Bring package level classes into the top level namespace
from masterful.version import __version__
from masterful.utils.register import register

# Pre-import the core packages.
import masterful.enums
import masterful.data
import masterful.architecture
import masterful.optimization
import masterful.regularization
import masterful.ssl
import masterful.training



""" Model architecture. """
from dataclasses import dataclass
import tensorflow as tf
from typing import Dict, Tuple


@dataclass
class ArchitectureParams:
  """Model architecture parameters.

  Parameters describing the model architecture used during training.
  These parameters describe structural attributes of the model,
  like input and output shapes, as well as higher order semantic
  parameters such as the task the model is attempting to solve.

  Args:
    task: A semantic description of the a model's predictions or groundtruth
      labels, such as single-label classification (CIFAR10/Imagenet) and
      detection (COCO).
    num_classes: The number of possible classes in the model prediction.
    ensemble_multiplier: The number of models to train as part of an ensemble.
      Defaults to 1, which disables ensembling and trains only a single model.
    custom_objects: Some TF/Keras models have custom objects (layers,
      initializers, etc). If so, in order to load/save/clone the model, you
      need to let TF/Keras know how to load/save them. A dictionary of class
      mappings is the typical way to do this.
    model_config: If the model is a Tensorflow Object Detection model,
      this is the configuration used to build the model.
    backbone_only: True if this model represents a backbone (feature extractor)
      only, without a classification or task specific head.
    input_shape: The input shape of image data the model expects,
      in the format (height, width, channels) if `input_channels_last=True`,
      otherwise (channels, height, width) if `input_channels_last=False`.
    input_range: The range of pixels in the input image space that
      the model is expecting.
    input_dtype: The image data type the model expects.
    input_channels_last: The ordering of the dimensions in the inputs.
      `input_channels_last=True` corresponds to inputs with shape
      (height, width, channels) while `input_channels_last=False`
      corresponds to inputs with shape (channels, height, width). Defaults
      to True.

    prediction_logits: True the model's output prediction is has no activation
      function. False means the model output is activated with a sigmoid
      or softmax layer.
    prediction_dtype: The dtype of the predictions.
    prediction_structure: The tensor structure of the model predictions.
    prediction_shape: The tensor shape of the model predictions. Only valid
      if the `prediction_structure` corresponds to `SINGLE_TENSOR`.
  """
  task: "masterful.enums.Task" = None
  num_classes: int = None
  ensemble_multiplier: int = 1
  custom_objects: Dict = None
  model_config: "object_detection.protos.model.DetectionModel" = None
  backbone_only: bool = False

  input_shape: Tuple = None
  input_range: "masterful.enums.ImageRange" = None
  input_dtype: tf.dtypes.DType = None
  input_channels_last: bool = True

  prediction_logits: bool = None
  prediction_dtype: tf.dtypes.DType = None
  prediction_structure: "masterful.enums.TensorStructure" = None
  prediction_shape: Tuple = None

  @property
  def input_width(self) -> int:
    """Gets the width of the model input."""
    if len(self.input_shape) > 2:
      return (self.input_shape[1]
              if self.input_channels_last else self.input_shape[2])
    else:
      return self.input_shape[1]

  @property
  def input_height(self) -> int:
    """Gets the height of the model input."""
    if len(self.input_shape) > 2:
      return (self.input_shape[0]
              if self.input_channels_last else self.input_shape[1])
    else:
      return self.input_shape[0]

  @property
  def input_channels(self) -> int:
    """Gets the number of channels in the model input."""
    if len(self.input_shape) > 2:
      return (self.input_shape[2]
              if self.input_channels_last else self.input_shape[0])
    else:
      return 0


def learn_architecture_params(
    model: tf.keras.Model,
    task: "masterful.enums.Task",
    input_range: "masterful.enums.ImageRange",
    num_classes: int,
    prediction_logits: bool,
    custom_objects: Dict = {},
    backbone_only: bool = False,
) -> ArchitectureParams:
  """Architecture parameter learner.

  Learns the parameters of the given model. Most attributes can
  be introspected from the model itself. Any attributes that
  cannot be introspected can be passed into the function, or set
  on the :class:`ArchitectureParams` after creation.

  Example:

  .. code-block:: python

    model: tf.keras.Model = ...
    architecture_params = masterful.architecture.learn_architecture_params(
        model=model,
        task=masterful.enums.Task.CLASSIFICATION,
        input_range=masterful.enums.ImageRange.ZERO_255,
        num_classes=10,
        prediction_logits=False)

  Args:
    model: The model to learn the parameters for.
    task: A semantic description of the a model's predictions or groundtruth
      labels, such as single-label classification (CIFAR10/Imagenet) and
      detection (COCO).
    input_range: The range of pixels in the input image space that
      the model is expecting.
    num_classes: The number of possible classes in the model prediction.
      If the model is a backbone only, then this should still be provided
      and represents the number of classes that should exist in in the
      classification head of the model.
    prediction_logits: True the model's output prediction is has no activation
      function. False means the model output is activated with a sigmoid
      or softmax layer.
    custom_objects: Some TF/Keras models have custom objects (layers,
      initializers, etc). If so, in order to load/save/clone the model, you
      need to let TF/Keras know how to load/save them. A dictionary of class
      mappings is the typical way to do this.
    backbone_only: True if this model represents a backbone (feature extractor)
      only, without a classification or task specific head. Defaults to False.

  Returns:
    An instance of :class:`ArchitectureParams` describing the attributes
    of the model.
  """
  raise RuntimeError(
      "Please call masterful.register() with your account ID and authorization key before using the API."
  )



import re
import setuptools
import masterful.version as version
from pkg_resources import get_distribution, DistributionNotFound

with open("README", "r") as fh:
  long_description = fh.read()

INSTALL_REQUIRES = [
    'protobuf',
    'pandas',
    'plotly',
    'inject',
    'tensorflow_addons',
    'psutil',
    'packaging',
    'typing_extensions',
    'tqdm',
    'memory_tempfile',
    'requests',
    'dataclasses',
    'albumentations',
    'pycocotools',
    'mful-gui-dev',
]

ALT_INSTALL_REQUIRES = {}


def check_alternative_installation(install_require,
                                   alternative_install_requires):
  """If some version version of alternative requirement installed, return alternative,
    else return main.
    """
  for alternative_install_require in alternative_install_requires:
    try:
      alternative_pkg_name = re.split(r"[!<>=]", alternative_install_require)[0]
      get_distribution(alternative_pkg_name)
      return str(alternative_install_require)
    except DistributionNotFound:
      continue

  return str(install_require)


def get_install_requirements(main_requires, alternative_requires):
  """Iterates over all install requires
    If an install require has an alternative option, check if this option is installed
    If that is the case, replace the install require by the alternative to not install dual package"""
  install_requires = []
  for main_require in main_requires:
    if main_require in alternative_requires:
      main_require = check_alternative_installation(
          main_require, alternative_requires.get(main_require))
    install_requires.append(main_require)

  return install_requires


INSTALL_REQUIRES = get_install_requirements(INSTALL_REQUIRES,
                                            ALT_INSTALL_REQUIRES)

setuptools.setup(
    name="mful-dev",
    version=version.__version__,
    author="Masterful AI",
    author_email="learn@masterfulai.com",
    description="Masterful AutoML Platform.",
    long_description_content_type="text/markdown",
    long_description=long_description,
    license_files=('LICENSE',),
    license='Copyright 2022, Masterful AI, Inc.',
    url="http://masterfulai.com",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
        "Operating System :: OS Independent",
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    install_requires=INSTALL_REQUIRES,
    python_requires=">=3.6",
)


"""Enumeration classes used in the Masterful API. """
import enum


@enum.unique
class Task(enum.Enum):
  """An enum to semantically specify a model's use case.

  Args:
    value: Overriden from :class:`enum.Enum`. Returns the member of this enum
      from the corresponding value.

  Attributes:
    CLASSIFICATION: Normal classification task like Alexnet on Imagenet.
    BINARY_CLASSIFICATION: Binary classification task.
    MULTILABEL_CLASSIFICATION: Multi-label classification task.
    DETECTION: Object detection (localization + classification) task.
    LOCALIZATION: Object localization task.
    SEMANTIC_SEGMENTATION: Semantic segmentation task.
    INSTANCE_SEGMENTATION: Instance segmentation task.
    KEYPOINT_DETECTION: Keypoint detection task.
  """
  CLASSIFICATION = "classification"
  BINARY_CLASSIFICATION = "binary_classification"
  MULTILABEL_CLASSIFICATION = "multilabel_classification"
  DETECTION = "detection"
  LOCALIZATION = "localization"
  SEMANTIC_SEGMENTATION = "semantic_segmentation"
  INSTANCE_SEGMENTATION = "instance_segmentation"
  KEYPOINT_DETECTION = "keypoint_detection"


@enum.unique
class ImageRange(enum.Enum):
  """An enum to model the image input ranges Masterful supports.

  ImageRange describes the range of the image pixel values.
  Common ranges include [0,255] and [0,1]. Some models also
  prefer to normalize the input data around the ImageNet
  mean and standard deviation.

  Args:
    value: Overriden from :class:`enum.Enum`. Returns the member of this enum
      from the corresponding value.

  Attributes:
    ZERO_ONE: Image range is [0,1].
    NEG_ONE_POS_ONE: Image range is [-1, 1].
    ZERO_255: Image range is [0,255].
    IMAGENET_CAFFE_BGR: Image is in BGR channel format, and each channel has been zero-centered around the Imagenet mean, without scaling.
    IMAGENET_TORCH: Image pixels were scaled to [0,1], then each channel was zero-centered around the Imagenet mean.
    CIFAR10_TORCH: Image pixels were scaled to [0,1], then each channel was zero-centered around the CIFAR10 mean.
    CIFAR100_TORCH: Image pixels were scaled to [0,1], then each channel was zero-centered around the CIFAR100 mean.
  """
  ZERO_ONE = "zero_one"
  NEG_ONE_POS_ONE = "neg_one_pos_one"
  ZERO_255 = "zero_255"
  IMAGENET_CAFFE_BGR = "imagenet_caffe_bgr"
  IMAGENET_TORCH = "imagenet_torch"
  CIFAR10_TORCH = "cifar10_torch"
  CIFAR100_TORCH = "cifar100_torch"


@enum.unique
class TensorStructure(enum.Enum):
  """An enum to specify tensor structures.

  A tensor structure is the physical data structure used to encapsulate
  individual tensors. This can either be a single tensor itself, in which
  case the single tensor is passed to each output of the model, or a tuple
  of tensors, in which case the number of tensors in the tuple must match exactly
  the number of outputs in the model.

  Attributes:
    SINGLE_TENSOR: Label or prediction consists of a single tensor, which is passed
      to all outputs of the model.

    TUPLE: Label or prediction consists of a tuple of tensors, which must map exactly
      to the number of outputs of a model.

    DICT: Label or prediction is a dictionary output tensors.
  """

  SINGLE_TENSOR = "single_tensor"
  TUPLE = "tuple"
  DICT = "dict"


@enum.unique
class BoundingBoxFormat(enum.Enum):
  """ Bounding box formats support by Masterful.

  Attributes:
    TENSORFLOW: The Tensorflow bounding box format is
      (ymin, xmin, ymax, xmax), normalized by the image dimensions
      into the range [0,1].
    VOC: The Pascal VOC bounding box format is
      (xmin, ymin, xmax, ymax) in pixel coordinates.
    COCO: The MSCOCO bounding box format is
      (xmin, ymin, width, height) in pixel coordinates.
    ALBUMENTATIONS: The Albumentations bounding box format is
      (xmin, ymin, xmax, ymax), normalized by the image dimension
      into the range [0,1]
    YOLO: The Yolo bounding box format is
      (x_center, y_center, width, height), normalized by the image
      dimensions into the range [0,1]
  """
  TENSORFLOW = 'tensorflow'
  VOC = 'voc'
  COCO = 'coco'
  ALBUMENTATIONS = 'albumentations'
  YOLO = 'yolo'


@enum.unique
class DatasetStatistics(enum.Enum):
  """An enum for datasets statistics required for Masterful's normalization."""
  CIFAR_10_MEAN = [0.4914, 0.4822, 0.4465]
  CIFAR_10_STD = [0.247, 0.243, 0.261]
  CIFAR_10_BGR_255 = [113.8575, 122.961, 125.307]
  CIFAR_100_MEAN = [0.5071, 0.4867, 0.4408]
  CIFAR_100_STD = [0.2675, 0.2565, 0.2761]
  CIFAR_100_BGR_255 = [129.3105, 124.1085, 112.404]
  IMAGENET_MEAN = [0.485, 0.456, 0.406]
  IMAGENET_STD = [0.229, 0.224, 0.225]
  IMAGENET_MEAN_BGR_255 = [103.939, 116.779, 123.68]



"""Regularization focused parameters and learning algorithms. """
import dataclasses
import numpy as np
import tensorflow as tf
from typing import Dict, Optional, Sequence, Tuple


@dataclasses.dataclass
class RegularizationParams:
  """Parameters controlling the regularization of the model during training.

  Regularization involves helping a model generalize
  to data it has not yet seen. Another way of saying this is that
  regularization is about fighting overfitting.

  Masterful supports a range of different augmentations in order
  to regularize the model during training. The core concept behind
  the Masterful augmentation strategy is "augmentation clustering",
  which groups structurally similar augmentations together
  (for example, color space augmentations are one group, spatial
  augmentations are another). In order to support this, we
  measure a range of different magnitudes for each cluster (the
  xxx_magnitude_table), the index of the magnitude selected for each
  cluster (xxx_cluster_to_index), and finally the optimal magnitude
  chosen for each cluster. Masterful automatically determines
  each cluster size and potential magnitude by measuring the
  "distance" each each augmentation moves the loss of the model.
  This is the "Distance Analysis" phase in the console output.

  Args:
    shuffle_buffer_size: The size of the shuffle buffer to use
      during training. If None or 0, the shuffle buffer size will
      be found automatically based on memory capacity.
    mirror: Specifies whether to apply mirror augmentation during
      training as part of the augmentation pipeline. Must be set to
      0.0 (off) or 1.0 (on).
    rot90: Specifies whether to apply a random discrete rotation during
      training as part of the augmentation pipeline. This will perform
      a random, discrete rotation of either  90, 180, or 270 degress. Must
      be set to either 0.0 (off) or 1.0 (on).
    rotate: Specifies whether to apply a random rotation during training
      as part of the augmentation pipeline. Must be an integer in the range
      [0,100], where 0 specifies no rotation, and 100 specifies maximum
      rotation (180 degrees).
    mixup: Controls Mixup label smoothing, as proposed in
      "mixup: Beyond Empirical Risk Minimization" (https://arxiv.org/abs/1710.09412).
      Must be a float in the range [0.0, 1.0].
    cutmix: Specifies whether to apply cutmix during training as part
      of the data augmentation pipeline. Cutmix was proposed in
      "CutMix: Regularization Strategy to Train Strong Classifiers with Localizable Features"
      (https://arxiv.org/abs/1905.04899) and must be a float in the
      range [0.0, 1.0].
    label_smoothing: Controls whether to use label smoothing regularization
      or not. Label Smoothing is a regularization technique that introduces
      noise for the labels. This accounts for the fact that datasets
      may have mistakes in them, so maximizing the likelihood of
      `log p(x|y)` directly can be harmful.
    hsv_cluster: The color space cluster to use during training. This
      is the index to use in order to find the magnitude of the cluster
      in the hsv_magnitude_table. The HSV cluster involves color,
      brightness, and hue jittering augmentations.
    hsv_cluster_to_index: Maps a cluster index to the index of the magnitude
      for that cluster in the hasv_magnitude_table.
    hsv_magnitude_table: A table of magnitudes to use for each cluster.
      The magnitude chosen is specified by `hsv_cluster`.
    contrast_cluster: The contrast cluster to use during training.
      This is the index to use in order to find the magnitude of
      the cluster in the hsv_magnitude_table. The contrast cluster
      involves (auto) contrast, solarize, posterize, and equalize
      augmentations.
    contrast_cluster_to_index: Maps a cluster index to the index of
      the magnitude for that cluster in the contrast_magnitude_table.
    contrast_magnitude_table: A table of magnitudes to use for each
      cluster. The magnitude chosen is specified by `contrast_cluster`.
    blur_cluster: The blur cluster to use during training. This
      is the index to use in order to find the magnitude of the cluster
      in the blur_magnitude_table. The blur contrast involves
      sharpening and desharpening augmentations.
    blur_cluster_to_index: Maps a cluster index to the index of
      the magnitude for that cluster in the blur_magnitude_table.
    blur_magnitude_table: A table of magnitudes to use for each
      cluster. The magnitude chosen is specified by `blur_cluster`.
    spatial_cluster: The spatial cluster to use during training.
      This is the index to use in order to find the magnitude of
      the cluster in the spatial_magnitude_table. The spatial cluster
      involves shear, translate, and zoom augmentations.
    spatial_cluster_to_index: Maps a cluster index to the index of
      the magnitude for that cluster in the spatial_magnitude_table.
    spatial_magnitude_table: A table of magnitudes to use for each
      cluster. The magnitude chosen is specified by `spatial_cluster`.
    synthetic_proportion: If synthetic data is used during training,
      the proportion of synthetic data to use from each dataset. Should
      be a list of floats in the range [0.0, 1.0].
  """
  shuffle_buffer_size: int = None

  mirror: float = None
  rot90: float = None
  rotate: int = None
  mixup: float = None
  cutmix: float = None
  label_smoothing: float = None

  hsv_cluster: int = None
  hsv_cluster_to_index: np.ndarray = None
  hsv_magnitude_table: np.ndarray = None

  contrast_cluster: int = None
  contrast_cluster_to_index: np.ndarray = None
  contrast_magnitude_table: np.ndarray = None

  blur_cluster: int = None
  blur_cluster_to_index: np.ndarray = None
  blur_magnitude_table: np.ndarray = None

  spatial_cluster: int = None
  spatial_cluster_to_index: np.ndarray = None
  spatial_magnitude_table: np.ndarray = None

  synthetic_proportion: Sequence[float] = None

  @staticmethod
  def load_from(path: str) -> 'RegularizationParams':
    """
    Loads the regularization params from the given path.

    Args:
      path: The full path to the regularization parameters to load.

    Returns:
      A new instance of :class:`masterful.regularization.RegularizationParams`.
    """
    raise RuntimeError(
        "Please call masterful.register() with your account ID and authorization key before using the API."
    )

  @staticmethod
  def save_to(params: 'RegularizationParams', path: str):
    """
    Saves the regularization params to the given path.

    Args:
      params: The instance of :class:`masterful.regularization.RegularizationParams` to save.
      path: The fully qualified path of the file to save the params into.
    """
    raise RuntimeError(
        "Please call masterful.register() with your account ID and authorization key before using the API."
    )


def learn_regularization_params(
    model: tf.keras.Model,
    model_params: "masterful.architecture.ArchitectureParams",
    optimization_params: "masterful.optimization.OptimizationParams",
    training_dataset: tf.data.Dataset,
    training_dataset_params: "masterful.data.DataParams",
    validation_dataset: Optional[tf.data.Dataset] = None,
    validation_dataset_params: Optional["masterful.data.DataParams"] = None,
    unlabeled_datasets: Optional[Sequence[Tuple[
        tf.data.Dataset, "masterful.data.DataParams"]]] = None,
    synthetic_datasets: Optional[Sequence[Tuple[
        tf.data.Dataset, "masterful.data.DataParams"]]] = None,
    **kwargs,
) -> RegularizationParams:
  """
  Learns the optimal set of regularization parameters for the model
  during training. This is dependent on having established the
  :class:`OptimizationParams` to be used during training before
  calling this function.

  This function can take awhile to complete. The expensive part
  of this algorithm is the augmentation learner. In general, this
  function will take around 1.5x the amount of time it takes to
  train your model, in order to learn the optimal set of augmentations
  to use as part of the regularization strategy.

  Example:

  .. code-block:: python

    model: tf.keras.Model = ...
    model_params: masterful.architecture.params.ArchitectureParams = ...
    optimization_params: masterful.optimization.params.OptimizationParams = ...
    training_dataset: tf.data.Dataset = ...
    training_dataset_params: masterful.data.params.DataParams = ...
    regularization_params = masterful.regularization.learn_regularization_params(
        model=model,
        model_params=model_params,
        optimization_params=optimization_params,
        training_dataset=training_dataset,
        training_dataset_params=training_dataset_params)

  Args:
    model: The model to learn the optimal set of regularization
      parameters for.
    model_params: The parameter of the model.
    optimization_params: The optimization parameters to use during training.
    training_dataset: The labeled dataset to use for learning.
    training_dataset_params: The parameters of the labeled dataset.
    validation_dataset: The optional validation dataset used to measure
      progress. If not validation dataset is specified, a small portion
      of the training dataset will be used as the validation set.
    validation_dataset_params: The parameters of the validation dataset.
    unlabeled_datasets: Optional sequence of unlabled datasets and their
      parameters, to use during training. If an unlabeled dataset is
      specified, then a set of algorithms must be specified in `ssl_params`
      otherwise this will have no effect.
    synthetic_datasets: Optional sequence of synthetic data and parameters
      to use during training. The amount of synthetic data used during
      training is controlled by
      :class:`masterful.regularization.RegularizationParams.synthetic_proportion`.

  Returns:
    An instance of :class:`RegularizationParams` describing the
    optimal regularization strategy to use during training.
  """
  raise RuntimeError(
      "Please call masterful.register() with your account ID and authorization key before using the API."
  )


parameters = type('', (), {})()

# These parameters were learned using a ResNet-20
# on CIFAR-10.
parameters.CIFAR10_SMALL = RegularizationParams(
    mirror=1.0,
    rot90=0.0,
    rotate=17,
    mixup=0.0,
    cutmix=0.0,
    label_smoothing=0,
    hsv_cluster=3,
    hsv_cluster_to_index=np.array(
        [[2, 4, 7, 11, 11, 11], [2, 5, 7, 11, 11, 11], [1, 2, 3, 4, 6, 11],
         [2, 4, 5, 6, 8, 11], [2, 2, 2, 3, 8, 11]],
        dtype=np.int32),
    hsv_magnitude_table=np.array([[0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100],
                                  [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100],
                                  [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100],
                                  [0, 20, 10, 30, 40, 50, 60, 70, 80, 90, 100],
                                  [0, 100, 90, 10, 80, 20, 30, 70, 60, 50, 40]],
                                 dtype=np.int32),
    contrast_cluster=3,
    contrast_cluster_to_index=np.array(
        [[3, 7, 11, 11, 11, 11], [1, 1, 1, 1, 4, 11], [4, 5, 6, 6, 9, 11],
         [1, 2, 5, 8, 11, 11], [1, 2, 4, 6, 11, 11], [2, 4, 5, 6, 7, 11]],
        dtype=np.int32),
    contrast_magnitude_table=np.array(
        [[0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100],
         [0, 10, 20, 50, 40, 30, 60, 70, 100, 80, 90],
         [10, 0, 20, 30, 40, 50, 60, 70, 80, 90, 100],
         [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100],
         [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100],
         [0, 20, 10, 30, 40, 50, 60, 70, 80, 90, 100]],
        dtype=np.int32),
    blur_cluster=3,
    blur_cluster_to_index=np.array(
        [[1, 2, 6, 11, 11, 11], [2, 8, 10, 11, 11, 11]], dtype=np.int32),
    blur_magnitude_table=np.array(
        [[0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100],
         [0, 70, 10, 60, 20, 80, 50, 30, 40, 90, 100]],
        dtype=np.int32),
    spatial_cluster=1,
    spatial_cluster_to_index=np.array(
        [[2, 5, 8, 11, 11, 11], [2, 4, 6, 7, 9, 11], [2, 5, 8, 11, 11, 11],
         [1, 2, 3, 4, 6, 11], [3, 5, 7, 9, 11, 11], [1, 1, 1, 2, 7, 11]],
        dtype=np.int32),
    spatial_magnitude_table=np.array(
        [[0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100],
         [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100],
         [0, 100, 10, 90, 20, 80, 30, 70, 60, 40, 50],
         [0, 10, 20, 30, 40, 50, 100, 90, 60, 80, 70],
         [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100],
         [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100]],
        dtype=np.int32),
    synthetic_proportion=[0.0])



""" Constants used by Masterful. """

import os
import typing


def _environment_override(env_variable: str,
                          default: str,
                          type: typing.Type = None) -> str:
  if type is not None:
    return (type(os.environ[env_variable])
            if env_variable in os.environ else default)
  else:
    return (os.environ[env_variable] if env_variable in os.environ else default)


MASTERFUL_ACCOUNT_ID = _environment_override('MASTERFUL_ACCOUNT_ID', '')
MASTERFUL_AUTHORIZATION_KEY = _environment_override(
    'MASTERFUL_AUTHORIZATION_KEY', '')
MASTERFUL_SESSION_ID = ''
MASTERFUL_USER_MAC_ADDRESS = ''

#
# DEV Url:  https://ti86ik5nx9.execute-api.us-east-1.amazonaws.com/dev
# PROD Url: https://kjsbngy38b.execute-api.us-east-1.amazonaws.com/prod/'
#
MASTERFUL_API_URL = _environment_override(
    'MASTERFUL_API_URL',
    'https://kjsbngy38b.execute-api.us-east-1.amazonaws.com/prod/')
MASTERFUL_LOAD = f'{MASTERFUL_API_URL}/load'
MASTERFUL_LOG_ENTRY = f'{MASTERFUL_API_URL}/log_entry'

MASTERFUL_VERBOSITY = _environment_override('MASTERFUL_VERBOSITY', 0, int)
